from dataclasses import dataclass

from _qwak_proto.qwak.feature_store.sources.batch_pb2 import (
    FileSystemConfiguration as ProtoFileSystemConfiguration,
    GcsServiceAccountImpersonation as ProtoGcsServiceAccountImpersonation,
    GcsUnauthenticated as ProtoGcsUnauthenticated,
)
from qwak.exceptions import QwakException
from qwak.feature_store.data_sources.batch.filesystem.base_config import (
    FileSystemConfiguration,
)


@dataclass
class GcpGcsServiceAccountImpersonation(FileSystemConfiguration):
    service_account_user: str

    def __post_init__(self):
        self._validate()

    def _validate(self):
        if not self.service_account_user or not self.service_account_user.strip():
            raise QwakException(
                "Service account user is mandatory for GCS service account impersonation, blanks are invalid"
            )

    def _to_proto(self):
        return ProtoFileSystemConfiguration(
            gcs_service_account_impersonation=ProtoGcsServiceAccountImpersonation(
                service_account_user=self.service_account_user
            )
        )

    @classmethod
    def _from_proto(cls, proto):
        return GcpGcsServiceAccountImpersonation(
            service_account_user=proto.service_account_user
        )


@dataclass
class GcpGcsUnauthenticated(FileSystemConfiguration):
    def _to_proto(self):
        return ProtoFileSystemConfiguration(
            gcs_unauthenticated=ProtoGcsUnauthenticated()
        )

    @classmethod
    def _from_proto(cls, proto):
        return GcpGcsUnauthenticated()



import dataclasses
import itertools
import re
from typing import Callable, Dict, List

from qwak.feature_store.offline import OfflineClient
from qwak.model.schema_entities import BaseFeature, Entity, FeatureStoreInput


def is_asterisk(feature: BaseFeature) -> bool:
    """
    Checks whether the given Feature is a FeatureStoreInput with an '*'
    Args:
        feature: input feature
    Return
    """
    p = "^[^.]+.\\*$"
    # if this is not a FeatureStoreInput, no need to check
    if isinstance(feature, FeatureStoreInput):
        return bool(re.match(p, feature.name))
    return False


def inflate_feature(feature: BaseFeature, offline_fs: OfflineClient) -> List:
    """
    Inflates a Feature into a list. if the feature contains an '*', a list
        a feature for each feature in the associate FeatureSet is returned. else -
        a list containing the original feature is returned.
    Args:
        feature: input feature
        offline_fs: OfflineFeatureStore to query
    Return:
        if an '*' exists, returns a List of FeatureStoreInuput for each feature in the feature set
        else returns a list containing the feature.

    """
    if is_asterisk(feature):
        featureset_name = feature.name.lower().split(".")[0]

        p = f"^{featureset_name}+.*$"
        feature_names = [
            feature.lower()
            for feature in offline_fs.get_columns_by_feature_set(featureset_name)
            if re.match(p, feature.lower())
        ]
        return [dataclasses.replace(feature, name=name) for name in feature_names]

    return [feature]


def unpack_asterisk_features(
    features: List[BaseFeature],
    feature_store_generator: Callable[[], OfflineClient] = lambda: OfflineClient(),
) -> List[BaseFeature]:
    """
    Handles features with an '*'.
    If a feature of type FeatureStoreInput that matches '<featureset_name>.*' exists,
    transforms this feature into the list of features actually present in the FeatureSet.
    Other features remain unchanged.
    the original order is maintained, and the internal order (inside a FeatureStoreInput
     that has an '*') is inferred from the order of columns returned from
     OfflineFeatureStore.get_columns_by_feature_set(...)
     Args:
         features: List of features
         feature_store_generator: A function that generates an instance of OfflineFeatureStore
                                  If no feature with an '*' is found, the function is not called.

    Returns:
        the unpacked features
    """
    asterisk_found = any([is_asterisk(f) for f in features])
    if asterisk_found:
        # at least 1 feature contains an asterisk, map each feature to
        # a list of features (features with no '*' are mapped to a list of size 1), then flatten it.
        offline_feature_store = feature_store_generator()
        inflated_features = [
            inflate_feature(feature, offline_feature_store) for feature in features
        ]
        return list(itertools.chain.from_iterable(inflated_features))

    return features


def unpack_asterisk_features_from_key_mapping(
    key_to_features: Dict[str, List[str]],
    feature_store_generator: Callable[[], OfflineClient] = lambda: OfflineClient(),
) -> dict:
    """
    Handles features with an '*'.
    If a feature name that matches '<featureset_name>.*' exists in one of the lists,
    transforms this feature into the list of feature names actually present in the FeatureSet.
    Other features remain unchanged.
    the original order is maintained, and the internal order (inside a feature
     that has an '*') is inferred from the order of columns returned from
     OfflineFeatureStore.get_columns_by_feature_set(...)
     Args:
         key_to_features: dictionary of entity keys to requested features (same as in the offline store)
             >>> key_to_features = {'uuid': ['user_purchases.*',
             >>>                             'user_purchases.avg_purchase_amount']}
         feature_store_generator: A function that generates an instance of OfflineFeatureStore
                                  If no feature with an '*' is found, the function is not called.
    Returns:
        the unpacked features
    """
    inflated_key_to_features = {}

    for entity, features in key_to_features.items():
        feature_inputs = [
            FeatureStoreInput(name=feature, entity=Entity(name=entity, type=str))
            for feature in features
        ]
        unpacked = unpack_asterisk_features(feature_inputs, feature_store_generator)
        feature_list = [feature.name for feature in unpacked]
        inflated_key_to_features[entity] = feature_list
    return inflated_key_to_features



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _OpsgenieRegion:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _OpsgenieRegionEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_OpsgenieRegion.ValueType], builtins.type):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    OPSGENIE_REGION_INVALID: _OpsgenieRegion.ValueType  # 0
    OPSGENIE_REGION_US: _OpsgenieRegion.ValueType  # 1
    OPSGENIE_REGION_EU: _OpsgenieRegion.ValueType  # 2

class OpsgenieRegion(_OpsgenieRegion, metaclass=_OpsgenieRegionEnumTypeWrapper): ...

OPSGENIE_REGION_INVALID: OpsgenieRegion.ValueType  # 0
OPSGENIE_REGION_US: OpsgenieRegion.ValueType  # 1
OPSGENIE_REGION_EU: OpsgenieRegion.ValueType  # 2
global___OpsgenieRegion = OpsgenieRegion

class OpsgenieIntegration(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ENCRYPTED_OPSGENIE_AUTH_CONFIG_FIELD_NUMBER: builtins.int
    OPSGENIE_SETTINGS_FIELD_NUMBER: builtins.int
    BASE_URL_FIELD_NUMBER: builtins.int
    OPSGEINIE_SYSTEM_SECRET_FIELD_NUMBER: builtins.int
    @property
    def encrypted_opsgenie_auth_config(self) -> global___OpsgenieAuthConfig:
        """Encrypted Opsgenie auth configuration"""
    @property
    def opsgenie_settings(self) -> global___OpsgenieSettings:
        """All Opsgenie settings"""
    base_url: builtins.str
    """Opsgenie base url. Determined by the Region"""
    @property
    def opsgeinie_system_secret(self) -> global___OpsgeinieSystemSecretDescriptor:
        """Opsgenie system secret descriptor"""
    def __init__(
        self,
        *,
        encrypted_opsgenie_auth_config: global___OpsgenieAuthConfig | None = ...,
        opsgenie_settings: global___OpsgenieSettings | None = ...,
        base_url: builtins.str = ...,
        opsgeinie_system_secret: global___OpsgeinieSystemSecretDescriptor | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth", b"auth", "encrypted_opsgenie_auth_config", b"encrypted_opsgenie_auth_config", "opsgeinie_system_secret", b"opsgeinie_system_secret", "opsgenie_settings", b"opsgenie_settings"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth", b"auth", "base_url", b"base_url", "encrypted_opsgenie_auth_config", b"encrypted_opsgenie_auth_config", "opsgeinie_system_secret", b"opsgeinie_system_secret", "opsgenie_settings", b"opsgenie_settings"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["auth", b"auth"]) -> typing_extensions.Literal["opsgeinie_system_secret"] | None: ...

global___OpsgenieIntegration = OpsgenieIntegration

class OpsgenieIntegrationSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    OPSGENIE_SETTINGS_FIELD_NUMBER: builtins.int
    OPSGENIE_AUTH_CONFIGURATION_FIELD_NUMBER: builtins.int
    @property
    def opsgenie_settings(self) -> global___OpsgenieSettings:
        """There used to be another field which was deprecated, hence, field numbers start from 2
        All Opsgenie settings
        """
    @property
    def opsgenie_auth_configuration(self) -> global___OpsgenieAuthConfig: ...
    def __init__(
        self,
        *,
        opsgenie_settings: global___OpsgenieSettings | None = ...,
        opsgenie_auth_configuration: global___OpsgenieAuthConfig | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth", b"auth", "opsgenie_auth_configuration", b"opsgenie_auth_configuration", "opsgenie_settings", b"opsgenie_settings"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth", b"auth", "opsgenie_auth_configuration", b"opsgenie_auth_configuration", "opsgenie_settings", b"opsgenie_settings"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["auth", b"auth"]) -> typing_extensions.Literal["opsgenie_auth_configuration"] | None: ...

global___OpsgenieIntegrationSpec = OpsgenieIntegrationSpec

class OpsgenieSettings(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REGION_FIELD_NUMBER: builtins.int
    region: global___OpsgenieRegion.ValueType
    """Opsgenie region"""
    def __init__(
        self,
        *,
        region: global___OpsgenieRegion.ValueType = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["region", b"region"]) -> None: ...

global___OpsgenieSettings = OpsgenieSettings

class OpsgenieAuthConfig(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    API_KEY_FIELD_NUMBER: builtins.int
    api_key: builtins.str
    def __init__(
        self,
        *,
        api_key: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["api_key", b"api_key"]) -> None: ...

global___OpsgenieAuthConfig = OpsgenieAuthConfig

class OpsgeinieSystemSecretDescriptor(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SECRET_NAME_FIELD_NUMBER: builtins.int
    API_KEY_SECRET_KEY_FIELD_NUMBER: builtins.int
    secret_name: builtins.str
    """Opsgenie system secret name"""
    api_key_secret_key: builtins.str
    """Opsgenie system secret api_key secret key"""
    def __init__(
        self,
        *,
        secret_name: builtins.str = ...,
        api_key_secret_key: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["api_key_secret_key", b"api_key_secret_key", "secret_name", b"secret_name"]) -> None: ...

global___OpsgeinieSystemSecretDescriptor = OpsgeinieSystemSecretDescriptor

class OpsgenieValidationSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    OPSGENIE_AUTH_CONFIG_FIELD_NUMBER: builtins.int
    REGION_FIELD_NUMBER: builtins.int
    @property
    def opsgenie_auth_config(self) -> global___OpsgenieAuthConfig:
        """Opsgenie auth config"""
    region: global___OpsgenieRegion.ValueType
    """Opsgenie region"""
    def __init__(
        self,
        *,
        opsgenie_auth_config: global___OpsgenieAuthConfig | None = ...,
        region: global___OpsgenieRegion.ValueType = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["opsgenie_auth_config", b"opsgenie_auth_config"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["opsgenie_auth_config", b"opsgenie_auth_config", "region", b"region"]) -> None: ...

global___OpsgenieValidationSpec = OpsgenieValidationSpec

class OpsgenieIntegrationOptions(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___OpsgenieIntegrationOptions = OpsgenieIntegrationOptions



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from _qwak_proto.qwak.builds import internal_builds_orchestrator_service_pb2 as qwak_dot_builds_dot_internal__builds__orchestrator__service__pb2


class InternalBuildsOrchestratorServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.BuildModel = channel.unary_unary(
                '/qwak.builds.orchestrator.InternalBuildsOrchestratorService/BuildModel',
                request_serializer=qwak_dot_builds_dot_internal__builds__orchestrator__service__pb2.InternalBuildModelRequest.SerializeToString,
                response_deserializer=qwak_dot_builds_dot_internal__builds__orchestrator__service__pb2.InternalBuildModelResponse.FromString,
                )


class InternalBuildsOrchestratorServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def BuildModel(self, request, context):
        """Build a serving model image
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_InternalBuildsOrchestratorServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'BuildModel': grpc.unary_unary_rpc_method_handler(
                    servicer.BuildModel,
                    request_deserializer=qwak_dot_builds_dot_internal__builds__orchestrator__service__pb2.InternalBuildModelRequest.FromString,
                    response_serializer=qwak_dot_builds_dot_internal__builds__orchestrator__service__pb2.InternalBuildModelResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'qwak.builds.orchestrator.InternalBuildsOrchestratorService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class InternalBuildsOrchestratorService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def BuildModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.builds.orchestrator.InternalBuildsOrchestratorService/BuildModel',
            qwak_dot_builds_dot_internal__builds__orchestrator__service__pb2.InternalBuildModelRequest.SerializeToString,
            qwak_dot_builds_dot_internal__builds__orchestrator__service__pb2.InternalBuildModelResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qwak/feature_store/sinks/sink.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from _qwak_proto.qwak.feature_store.sources import streaming_pb2 as qwak_dot_feature__store_dot_sources_dot_streaming__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#qwak/feature_store/sinks/sink.proto\x12\x18qwak.feature.store.sinks\x1a*qwak/feature_store/sources/streaming.proto\"\xb3\x01\n\rStreamingSink\x12L\n\x10\x61ttachment_point\x18\x01 \x01(\x0b\x32\x32.qwak.feature.store.sinks.StreamingAttachmentPoint\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x39\n\nkafka_sink\x18\x02 \x01(\x0b\x32#.qwak.feature.store.sinks.KafkaSinkH\x00\x42\x0b\n\tsink_type\"\xb9\x01\n\tKafkaSink\x12\r\n\x05topic\x18\x01 \x01(\t\x12\x19\n\x11\x62ootstrap_servers\x18\x02 \x01(\t\x12\x41\n\x0emessage_format\x18\x03 \x01(\x0e\x32).qwak.feature.store.sources.MessageFormat\x12?\n\x0b\x61uth_config\x18\x04 \x01(\x0b\x32*.qwak.feature.store.sources.Authentication\" \n\x1eOnlineStreamingAttachmentPoint\"!\n\x1fOfflineStreamingAttachmentPoint\"\x83\x02\n\x18StreamingAttachmentPoint\x12\x65\n!online_streaming_attachment_point\x18\x01 \x01(\x0b\x32\x38.qwak.feature.store.sinks.OnlineStreamingAttachmentPointH\x00\x12g\n\"offline_streaming_attachment_point\x18\x02 \x01(\x0b\x32\x39.qwak.feature.store.sinks.OfflineStreamingAttachmentPointH\x00\x42\x17\n\x15\x61ttachment_point_typeBR\n#com.qwak.ai.feature.store.sinks.apiP\x01Z)qwak/featurestore/sinks;featurestoresinksb\x06proto3')



_STREAMINGSINK = DESCRIPTOR.message_types_by_name['StreamingSink']
_KAFKASINK = DESCRIPTOR.message_types_by_name['KafkaSink']
_ONLINESTREAMINGATTACHMENTPOINT = DESCRIPTOR.message_types_by_name['OnlineStreamingAttachmentPoint']
_OFFLINESTREAMINGATTACHMENTPOINT = DESCRIPTOR.message_types_by_name['OfflineStreamingAttachmentPoint']
_STREAMINGATTACHMENTPOINT = DESCRIPTOR.message_types_by_name['StreamingAttachmentPoint']
StreamingSink = _reflection.GeneratedProtocolMessageType('StreamingSink', (_message.Message,), {
  'DESCRIPTOR' : _STREAMINGSINK,
  '__module__' : 'qwak.feature_store.sinks.sink_pb2'
  # @@protoc_insertion_point(class_scope:qwak.feature.store.sinks.StreamingSink)
  })
_sym_db.RegisterMessage(StreamingSink)

KafkaSink = _reflection.GeneratedProtocolMessageType('KafkaSink', (_message.Message,), {
  'DESCRIPTOR' : _KAFKASINK,
  '__module__' : 'qwak.feature_store.sinks.sink_pb2'
  # @@protoc_insertion_point(class_scope:qwak.feature.store.sinks.KafkaSink)
  })
_sym_db.RegisterMessage(KafkaSink)

OnlineStreamingAttachmentPoint = _reflection.GeneratedProtocolMessageType('OnlineStreamingAttachmentPoint', (_message.Message,), {
  'DESCRIPTOR' : _ONLINESTREAMINGATTACHMENTPOINT,
  '__module__' : 'qwak.feature_store.sinks.sink_pb2'
  # @@protoc_insertion_point(class_scope:qwak.feature.store.sinks.OnlineStreamingAttachmentPoint)
  })
_sym_db.RegisterMessage(OnlineStreamingAttachmentPoint)

OfflineStreamingAttachmentPoint = _reflection.GeneratedProtocolMessageType('OfflineStreamingAttachmentPoint', (_message.Message,), {
  'DESCRIPTOR' : _OFFLINESTREAMINGATTACHMENTPOINT,
  '__module__' : 'qwak.feature_store.sinks.sink_pb2'
  # @@protoc_insertion_point(class_scope:qwak.feature.store.sinks.OfflineStreamingAttachmentPoint)
  })
_sym_db.RegisterMessage(OfflineStreamingAttachmentPoint)

StreamingAttachmentPoint = _reflection.GeneratedProtocolMessageType('StreamingAttachmentPoint', (_message.Message,), {
  'DESCRIPTOR' : _STREAMINGATTACHMENTPOINT,
  '__module__' : 'qwak.feature_store.sinks.sink_pb2'
  # @@protoc_insertion_point(class_scope:qwak.feature.store.sinks.StreamingAttachmentPoint)
  })
_sym_db.RegisterMessage(StreamingAttachmentPoint)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n#com.qwak.ai.feature.store.sinks.apiP\001Z)qwak/featurestore/sinks;featurestoresinks'
  _STREAMINGSINK._serialized_start=110
  _STREAMINGSINK._serialized_end=289
  _KAFKASINK._serialized_start=292
  _KAFKASINK._serialized_end=477
  _ONLINESTREAMINGATTACHMENTPOINT._serialized_start=479
  _ONLINESTREAMINGATTACHMENTPOINT._serialized_end=511
  _OFFLINESTREAMINGATTACHMENTPOINT._serialized_start=513
  _OFFLINESTREAMINGATTACHMENTPOINT._serialized_end=546
  _STREAMINGATTACHMENTPOINT._serialized_start=549
  _STREAMINGATTACHMENTPOINT._serialized_end=808
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from _qwak_proto.qwak.execution.v1.jobs import job_service_pb2 as qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2


class JobServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ApplyJobRecord = channel.unary_unary(
                '/qwak.feature.store.execution.v1.jobs.JobService/ApplyJobRecord',
                request_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ApplyJobRecordRequest.SerializeToString,
                response_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ApplyJobRecordResponse.FromString,
                )
        self.InitPagination = channel.unary_unary(
                '/qwak.feature.store.execution.v1.jobs.JobService/InitPagination',
                request_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.InitPaginationRequest.SerializeToString,
                response_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.InitPaginationResponse.FromString,
                )
        self.ListJobs = channel.unary_unary(
                '/qwak.feature.store.execution.v1.jobs.JobService/ListJobs',
                request_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ListJobsRequest.SerializeToString,
                response_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ListJobsResponse.FromString,
                )
        self.InitPaginationByExecutionId = channel.unary_unary(
                '/qwak.feature.store.execution.v1.jobs.JobService/InitPaginationByExecutionId',
                request_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.InitPaginationByExecutionIdRequest.SerializeToString,
                response_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.InitPaginationByExecutionIdResponse.FromString,
                )
        self.ListJobsByExecutionId = channel.unary_unary(
                '/qwak.feature.store.execution.v1.jobs.JobService/ListJobsByExecutionId',
                request_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ListJobsByExecutionIdRequest.SerializeToString,
                response_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ListJobsByExecutionIdResponse.FromString,
                )
        self.DeleteFeaturesetJobs = channel.unary_unary(
                '/qwak.feature.store.execution.v1.jobs.JobService/DeleteFeaturesetJobs',
                request_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.DeleteFeaturesetJobsRequest.SerializeToString,
                response_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.DeleteFeaturesetJobsResponse.FromString,
                )
        self.GetLatestSuccessfulJobByFeaturesetId = channel.unary_unary(
                '/qwak.feature.store.execution.v1.jobs.JobService/GetLatestSuccessfulJobByFeaturesetId',
                request_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetLatestSuccessfulJobByFeaturesetIdRequest.SerializeToString,
                response_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetLatestSuccessfulJobByFeaturesetIdResponse.FromString,
                )
        self.GetAllFeaturesetLatestSuccessfulJobs = channel.unary_unary(
                '/qwak.feature.store.execution.v1.jobs.JobService/GetAllFeaturesetLatestSuccessfulJobs',
                request_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetAllFeaturesetLatestSuccessfulJobsRequest.SerializeToString,
                response_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetAllFeaturesetLatestSuccessfulJobsResponse.FromString,
                )
        self.GetAllFeaturesetJobsSummaries = channel.unary_unary(
                '/qwak.feature.store.execution.v1.jobs.JobService/GetAllFeaturesetJobsSummaries',
                request_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetAllFeaturesetJobsSummariesRequest.SerializeToString,
                response_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetAllFeaturesetJobsSummariesResponse.FromString,
                )
        self.GetJobsSummaryByExecutionId = channel.unary_unary(
                '/qwak.feature.store.execution.v1.jobs.JobService/GetJobsSummaryByExecutionId',
                request_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetJobsSummaryByExecutionIdRequest.SerializeToString,
                response_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetJobsSummaryByExecutionIdResponse.FromString,
                )
        self.GetJobSummariesByExecutionIds = channel.unary_unary(
                '/qwak.feature.store.execution.v1.jobs.JobService/GetJobSummariesByExecutionIds',
                request_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetJobSummariesByExecutionIdsRequest.SerializeToString,
                response_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetJobSummariesByExecutionIdsResponse.FromString,
                )


class JobServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def ApplyJobRecord(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def InitPagination(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListJobs(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def InitPaginationByExecutionId(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListJobsByExecutionId(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteFeaturesetJobs(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetLatestSuccessfulJobByFeaturesetId(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAllFeaturesetLatestSuccessfulJobs(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAllFeaturesetJobsSummaries(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobsSummaryByExecutionId(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobSummariesByExecutionIds(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_JobServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ApplyJobRecord': grpc.unary_unary_rpc_method_handler(
                    servicer.ApplyJobRecord,
                    request_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ApplyJobRecordRequest.FromString,
                    response_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ApplyJobRecordResponse.SerializeToString,
            ),
            'InitPagination': grpc.unary_unary_rpc_method_handler(
                    servicer.InitPagination,
                    request_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.InitPaginationRequest.FromString,
                    response_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.InitPaginationResponse.SerializeToString,
            ),
            'ListJobs': grpc.unary_unary_rpc_method_handler(
                    servicer.ListJobs,
                    request_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ListJobsRequest.FromString,
                    response_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ListJobsResponse.SerializeToString,
            ),
            'InitPaginationByExecutionId': grpc.unary_unary_rpc_method_handler(
                    servicer.InitPaginationByExecutionId,
                    request_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.InitPaginationByExecutionIdRequest.FromString,
                    response_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.InitPaginationByExecutionIdResponse.SerializeToString,
            ),
            'ListJobsByExecutionId': grpc.unary_unary_rpc_method_handler(
                    servicer.ListJobsByExecutionId,
                    request_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ListJobsByExecutionIdRequest.FromString,
                    response_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ListJobsByExecutionIdResponse.SerializeToString,
            ),
            'DeleteFeaturesetJobs': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteFeaturesetJobs,
                    request_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.DeleteFeaturesetJobsRequest.FromString,
                    response_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.DeleteFeaturesetJobsResponse.SerializeToString,
            ),
            'GetLatestSuccessfulJobByFeaturesetId': grpc.unary_unary_rpc_method_handler(
                    servicer.GetLatestSuccessfulJobByFeaturesetId,
                    request_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetLatestSuccessfulJobByFeaturesetIdRequest.FromString,
                    response_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetLatestSuccessfulJobByFeaturesetIdResponse.SerializeToString,
            ),
            'GetAllFeaturesetLatestSuccessfulJobs': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAllFeaturesetLatestSuccessfulJobs,
                    request_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetAllFeaturesetLatestSuccessfulJobsRequest.FromString,
                    response_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetAllFeaturesetLatestSuccessfulJobsResponse.SerializeToString,
            ),
            'GetAllFeaturesetJobsSummaries': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAllFeaturesetJobsSummaries,
                    request_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetAllFeaturesetJobsSummariesRequest.FromString,
                    response_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetAllFeaturesetJobsSummariesResponse.SerializeToString,
            ),
            'GetJobsSummaryByExecutionId': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJobsSummaryByExecutionId,
                    request_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetJobsSummaryByExecutionIdRequest.FromString,
                    response_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetJobsSummaryByExecutionIdResponse.SerializeToString,
            ),
            'GetJobSummariesByExecutionIds': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJobSummariesByExecutionIds,
                    request_deserializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetJobSummariesByExecutionIdsRequest.FromString,
                    response_serializer=qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetJobSummariesByExecutionIdsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'qwak.feature.store.execution.v1.jobs.JobService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class JobService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def ApplyJobRecord(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.feature.store.execution.v1.jobs.JobService/ApplyJobRecord',
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ApplyJobRecordRequest.SerializeToString,
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ApplyJobRecordResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def InitPagination(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.feature.store.execution.v1.jobs.JobService/InitPagination',
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.InitPaginationRequest.SerializeToString,
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.InitPaginationResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListJobs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.feature.store.execution.v1.jobs.JobService/ListJobs',
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ListJobsRequest.SerializeToString,
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ListJobsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def InitPaginationByExecutionId(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.feature.store.execution.v1.jobs.JobService/InitPaginationByExecutionId',
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.InitPaginationByExecutionIdRequest.SerializeToString,
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.InitPaginationByExecutionIdResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListJobsByExecutionId(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.feature.store.execution.v1.jobs.JobService/ListJobsByExecutionId',
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ListJobsByExecutionIdRequest.SerializeToString,
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.ListJobsByExecutionIdResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteFeaturesetJobs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.feature.store.execution.v1.jobs.JobService/DeleteFeaturesetJobs',
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.DeleteFeaturesetJobsRequest.SerializeToString,
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.DeleteFeaturesetJobsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetLatestSuccessfulJobByFeaturesetId(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.feature.store.execution.v1.jobs.JobService/GetLatestSuccessfulJobByFeaturesetId',
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetLatestSuccessfulJobByFeaturesetIdRequest.SerializeToString,
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetLatestSuccessfulJobByFeaturesetIdResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetAllFeaturesetLatestSuccessfulJobs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.feature.store.execution.v1.jobs.JobService/GetAllFeaturesetLatestSuccessfulJobs',
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetAllFeaturesetLatestSuccessfulJobsRequest.SerializeToString,
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetAllFeaturesetLatestSuccessfulJobsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetAllFeaturesetJobsSummaries(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.feature.store.execution.v1.jobs.JobService/GetAllFeaturesetJobsSummaries',
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetAllFeaturesetJobsSummariesRequest.SerializeToString,
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetAllFeaturesetJobsSummariesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetJobsSummaryByExecutionId(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.feature.store.execution.v1.jobs.JobService/GetJobsSummaryByExecutionId',
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetJobsSummaryByExecutionIdRequest.SerializeToString,
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetJobsSummaryByExecutionIdResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetJobSummariesByExecutionIds(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.feature.store.execution.v1.jobs.JobService/GetJobSummariesByExecutionIds',
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetJobSummariesByExecutionIdsRequest.SerializeToString,
            qwak_dot_execution_dot_v1_dot_jobs_dot_job__service__pb2.GetJobSummariesByExecutionIdsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from _qwak_proto.qwak.feature_store.features import feature_set_state_service_pb2 as qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2


class FeatureSetStateServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetFeatureSetStateById = channel.unary_unary(
                '/qwak.feature.store.features.FeatureSetStateService/GetFeatureSetStateById',
                request_serializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStateByIdV1Request.SerializeToString,
                response_deserializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStateByIdV1Response.FromString,
                )
        self.GetFeatureSetStateByName = channel.unary_unary(
                '/qwak.feature.store.features.FeatureSetStateService/GetFeatureSetStateByName',
                request_serializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStateByNameV1Request.SerializeToString,
                response_deserializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStateByNameV1Response.FromString,
                )
        self.GetFeatureSetStatesList = channel.unary_unary(
                '/qwak.feature.store.features.FeatureSetStateService/GetFeatureSetStatesList',
                request_serializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStatesListV1Request.SerializeToString,
                response_deserializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStatesListV1Response.FromString,
                )
        self.UpdateFeatureSetState = channel.unary_unary(
                '/qwak.feature.store.features.FeatureSetStateService/UpdateFeatureSetState',
                request_serializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.UpdateFeatureSetStateV1Request.SerializeToString,
                response_deserializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.UpdateFeatureSetStateV1Response.FromString,
                )
        self.DeleteFeatureSetState = channel.unary_unary(
                '/qwak.feature.store.features.FeatureSetStateService/DeleteFeatureSetState',
                request_serializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.DeleteFeatureSetStateV1Request.SerializeToString,
                response_deserializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.DeleteFeatureSetStateV1Response.FromString,
                )


class FeatureSetStateServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetFeatureSetStateById(self, request, context):
        """Get feature state by feature id
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetFeatureSetStateByName(self, request, context):
        """Get feature state by feature name
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetFeatureSetStatesList(self, request, context):
        """Get list of feature state
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateFeatureSetState(self, request, context):
        """Update feature set state
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteFeatureSetState(self, request, context):
        """Delete feature set state
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_FeatureSetStateServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetFeatureSetStateById': grpc.unary_unary_rpc_method_handler(
                    servicer.GetFeatureSetStateById,
                    request_deserializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStateByIdV1Request.FromString,
                    response_serializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStateByIdV1Response.SerializeToString,
            ),
            'GetFeatureSetStateByName': grpc.unary_unary_rpc_method_handler(
                    servicer.GetFeatureSetStateByName,
                    request_deserializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStateByNameV1Request.FromString,
                    response_serializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStateByNameV1Response.SerializeToString,
            ),
            'GetFeatureSetStatesList': grpc.unary_unary_rpc_method_handler(
                    servicer.GetFeatureSetStatesList,
                    request_deserializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStatesListV1Request.FromString,
                    response_serializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStatesListV1Response.SerializeToString,
            ),
            'UpdateFeatureSetState': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateFeatureSetState,
                    request_deserializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.UpdateFeatureSetStateV1Request.FromString,
                    response_serializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.UpdateFeatureSetStateV1Response.SerializeToString,
            ),
            'DeleteFeatureSetState': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteFeatureSetState,
                    request_deserializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.DeleteFeatureSetStateV1Request.FromString,
                    response_serializer=qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.DeleteFeatureSetStateV1Response.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'qwak.feature.store.features.FeatureSetStateService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class FeatureSetStateService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetFeatureSetStateById(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.feature.store.features.FeatureSetStateService/GetFeatureSetStateById',
            qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStateByIdV1Request.SerializeToString,
            qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStateByIdV1Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetFeatureSetStateByName(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.feature.store.features.FeatureSetStateService/GetFeatureSetStateByName',
            qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStateByNameV1Request.SerializeToString,
            qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStateByNameV1Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetFeatureSetStatesList(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.feature.store.features.FeatureSetStateService/GetFeatureSetStatesList',
            qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStatesListV1Request.SerializeToString,
            qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.GetFeatureSetStatesListV1Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateFeatureSetState(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.feature.store.features.FeatureSetStateService/UpdateFeatureSetState',
            qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.UpdateFeatureSetStateV1Request.SerializeToString,
            qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.UpdateFeatureSetStateV1Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteFeatureSetState(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.feature.store.features.FeatureSetStateService/DeleteFeatureSetState',
            qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.DeleteFeatureSetStateV1Request.SerializeToString,
            qwak_dot_feature__store_dot_features_dot_feature__set__state__service__pb2.DeleteFeatureSetStateV1Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qwak/vectors/v1/collection/event/collection_event.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n7qwak/vectors/v1/collection/event/collection_event.proto\x12%qwak.vector.store.v1.collection.event\x1a\x1fgoogle/protobuf/timestamp.proto\"t\n\x0f\x43ollectionEvent\x12\x15\n\rcollection_id\x18\x01 \x01(\t\x12\x34\n\x10last_modified_at\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0cretry_number\x18\x03 \x01(\x05\x42L\n\'com.qwak.ai.vectors.v1.collection.eventP\x01Z\x1fqwak/vector_management/v1;eventb\x06proto3')



_COLLECTIONEVENT = DESCRIPTOR.message_types_by_name['CollectionEvent']
CollectionEvent = _reflection.GeneratedProtocolMessageType('CollectionEvent', (_message.Message,), {
  'DESCRIPTOR' : _COLLECTIONEVENT,
  '__module__' : 'qwak.vectors.v1.collection.event.collection_event_pb2'
  # @@protoc_insertion_point(class_scope:qwak.vector.store.v1.collection.event.CollectionEvent)
  })
_sym_db.RegisterMessage(CollectionEvent)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\'com.qwak.ai.vectors.v1.collection.eventP\001Z\037qwak/vector_management/v1;event'
  _COLLECTIONEVENT._serialized_start=131
  _COLLECTIONEVENT._serialized_end=247
# @@protoc_insertion_point(module_scope)



from abc import ABC
from dataclasses import dataclass, field

from qwak.llmops.prompt.value import PromptValue


@dataclass
class BaseMessage(PromptValue, ABC):
    content: str
    role_name: str = field(
        init=False,
    )


class AIMessage(BaseMessage):
    role_name: str = "ai"


class HumanMessage(BaseMessage):
    role_name: str = "human"


class SystemMessage(BaseMessage):
    role_name: str = "system"



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from _qwak_proto.qwak.monitoring.v0 import alerting_channel_sync_service_pb2 as qwak_dot_monitoring_dot_v0_dot_alerting__channel__sync__service__pb2


class AlertingChannelSyncServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ListAlertingChannelsForEdgeCluster = channel.unary_unary(
                '/qwak.monitoring.v0.AlertingChannelSyncService/ListAlertingChannelsForEdgeCluster',
                request_serializer=qwak_dot_monitoring_dot_v0_dot_alerting__channel__sync__service__pb2.ListAlertingChannelsForEdgeClusterRequest.SerializeToString,
                response_deserializer=qwak_dot_monitoring_dot_v0_dot_alerting__channel__sync__service__pb2.ListAlertingChannelsForEdgeClusterResponse.FromString,
                )
        self.ListAlertingChannelsForCoreCluster = channel.unary_unary(
                '/qwak.monitoring.v0.AlertingChannelSyncService/ListAlertingChannelsForCoreCluster',
                request_serializer=qwak_dot_monitoring_dot_v0_dot_alerting__channel__sync__service__pb2.ListAlertingChannelsForCoreClusterRequest.SerializeToString,
                response_deserializer=qwak_dot_monitoring_dot_v0_dot_alerting__channel__sync__service__pb2.ListAlertingChannelsForCoreClusterResponse.FromString,
                )


class AlertingChannelSyncServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def ListAlertingChannelsForEdgeCluster(self, request, context):
        """List alerting channels for edge channels
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListAlertingChannelsForCoreCluster(self, request, context):
        """List alerting channels for core channels
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AlertingChannelSyncServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ListAlertingChannelsForEdgeCluster': grpc.unary_unary_rpc_method_handler(
                    servicer.ListAlertingChannelsForEdgeCluster,
                    request_deserializer=qwak_dot_monitoring_dot_v0_dot_alerting__channel__sync__service__pb2.ListAlertingChannelsForEdgeClusterRequest.FromString,
                    response_serializer=qwak_dot_monitoring_dot_v0_dot_alerting__channel__sync__service__pb2.ListAlertingChannelsForEdgeClusterResponse.SerializeToString,
            ),
            'ListAlertingChannelsForCoreCluster': grpc.unary_unary_rpc_method_handler(
                    servicer.ListAlertingChannelsForCoreCluster,
                    request_deserializer=qwak_dot_monitoring_dot_v0_dot_alerting__channel__sync__service__pb2.ListAlertingChannelsForCoreClusterRequest.FromString,
                    response_serializer=qwak_dot_monitoring_dot_v0_dot_alerting__channel__sync__service__pb2.ListAlertingChannelsForCoreClusterResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'qwak.monitoring.v0.AlertingChannelSyncService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class AlertingChannelSyncService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def ListAlertingChannelsForEdgeCluster(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.monitoring.v0.AlertingChannelSyncService/ListAlertingChannelsForEdgeCluster',
            qwak_dot_monitoring_dot_v0_dot_alerting__channel__sync__service__pb2.ListAlertingChannelsForEdgeClusterRequest.SerializeToString,
            qwak_dot_monitoring_dot_v0_dot_alerting__channel__sync__service__pb2.ListAlertingChannelsForEdgeClusterResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListAlertingChannelsForCoreCluster(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/qwak.monitoring.v0.AlertingChannelSyncService/ListAlertingChannelsForCoreCluster',
            qwak_dot_monitoring_dot_v0_dot_alerting__channel__sync__service__pb2.ListAlertingChannelsForCoreClusterRequest.SerializeToString,
            qwak_dot_monitoring_dot_v0_dot_alerting__channel__sync__service__pb2.ListAlertingChannelsForCoreClusterResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import uuid
from abc import ABC, abstractmethod

import pandas as pd


class BaseQueryEngine(ABC):
    @abstractmethod
    def upload_table(self, df: pd.DataFrame):
        pass

    @abstractmethod
    def cleanup(self):
        pass

    @abstractmethod
    def run_query(self, query: str):
        pass

    @abstractmethod
    def read_pandas_from_query(self, query: str, parse_dates=None):
        pass

    @staticmethod
    @abstractmethod
    def get_quotes():
        pass

    class JoinTableSpec:
        def __init__(self, join_tables_db_name: str, quotes: str):
            self.table_name = str(uuid.uuid4())
            self.join_table_full_path = f"{quotes}{join_tables_db_name}{quotes}.{quotes}{self.table_name}{quotes}"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qwak/fitness_service/fitness_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from _qwak_proto.qwak.fitness_service import fitness_pb2 as qwak_dot_fitness__service_dot_fitness__pb2
from _qwak_proto.qwak.fitness_service import status_pb2 as qwak_dot_fitness__service_dot_status__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n*qwak/fitness_service/fitness_service.proto\x12\x14qwak.fitness.service\x1a\"qwak/fitness_service/fitness.proto\x1a!qwak/fitness_service/status.proto\"H\n\x11\x42uildModelRequest\x12\x33\n\nbuild_spec\x18\x01 \x01(\x0b\x32\x1f.qwak.fitness.service.BuildSpec\"\x14\n\x12\x42uildModelResponse\"+\n\x17\x43\x61ncelBuildModelRequest\x12\x10\n\x08\x62uild_id\x18\x01 \x01(\t\"\x1a\n\x18\x43\x61ncelBuildModelResponse\"2\n\x1eGetRemoteBuildingStatusRequest\x12\x10\n\x08\x62uild_id\x18\x01 \x01(\t\"]\n\x1fGetRemoteBuildingStatusResponse\x12:\n\x06status\x18\x01 \x01(\x0b\x32*.qwak.fitness.service.RemoteBuildingStatus\"4\n CleanRemoteBuildResourcesRequest\x12\x10\n\x08\x62uild_id\x18\x01 \x01(\t\"#\n!CleanRemoteBuildResourcesResponse2\xfc\x03\n\x0e\x46itnessService\x12_\n\nBuildModel\x12\'.qwak.fitness.service.BuildModelRequest\x1a(.qwak.fitness.service.BuildModelResponse\x12q\n\x10\x43\x61ncelBuildModel\x12-.qwak.fitness.service.CancelBuildModelRequest\x1a..qwak.fitness.service.CancelBuildModelResponse\x12\x86\x01\n\x17GetRemoteBuildingStatus\x12\x34.qwak.fitness.service.GetRemoteBuildingStatusRequest\x1a\x35.qwak.fitness.service.GetRemoteBuildingStatusResponse\x12\x8c\x01\n\x19\x43leanRemoteBuildResources\x12\x36.qwak.fitness.service.CleanRemoteBuildResourcesRequest\x1a\x37.qwak.fitness.service.CleanRemoteBuildResourcesResponseB#\n\x1f\x63om.qwak.ai.fitness.service.apiP\x01\x62\x06proto3')



_BUILDMODELREQUEST = DESCRIPTOR.message_types_by_name['BuildModelRequest']
_BUILDMODELRESPONSE = DESCRIPTOR.message_types_by_name['BuildModelResponse']
_CANCELBUILDMODELREQUEST = DESCRIPTOR.message_types_by_name['CancelBuildModelRequest']
_CANCELBUILDMODELRESPONSE = DESCRIPTOR.message_types_by_name['CancelBuildModelResponse']
_GETREMOTEBUILDINGSTATUSREQUEST = DESCRIPTOR.message_types_by_name['GetRemoteBuildingStatusRequest']
_GETREMOTEBUILDINGSTATUSRESPONSE = DESCRIPTOR.message_types_by_name['GetRemoteBuildingStatusResponse']
_CLEANREMOTEBUILDRESOURCESREQUEST = DESCRIPTOR.message_types_by_name['CleanRemoteBuildResourcesRequest']
_CLEANREMOTEBUILDRESOURCESRESPONSE = DESCRIPTOR.message_types_by_name['CleanRemoteBuildResourcesResponse']
BuildModelRequest = _reflection.GeneratedProtocolMessageType('BuildModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _BUILDMODELREQUEST,
  '__module__' : 'qwak.fitness_service.fitness_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.fitness.service.BuildModelRequest)
  })
_sym_db.RegisterMessage(BuildModelRequest)

BuildModelResponse = _reflection.GeneratedProtocolMessageType('BuildModelResponse', (_message.Message,), {
  'DESCRIPTOR' : _BUILDMODELRESPONSE,
  '__module__' : 'qwak.fitness_service.fitness_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.fitness.service.BuildModelResponse)
  })
_sym_db.RegisterMessage(BuildModelResponse)

CancelBuildModelRequest = _reflection.GeneratedProtocolMessageType('CancelBuildModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _CANCELBUILDMODELREQUEST,
  '__module__' : 'qwak.fitness_service.fitness_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.fitness.service.CancelBuildModelRequest)
  })
_sym_db.RegisterMessage(CancelBuildModelRequest)

CancelBuildModelResponse = _reflection.GeneratedProtocolMessageType('CancelBuildModelResponse', (_message.Message,), {
  'DESCRIPTOR' : _CANCELBUILDMODELRESPONSE,
  '__module__' : 'qwak.fitness_service.fitness_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.fitness.service.CancelBuildModelResponse)
  })
_sym_db.RegisterMessage(CancelBuildModelResponse)

GetRemoteBuildingStatusRequest = _reflection.GeneratedProtocolMessageType('GetRemoteBuildingStatusRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETREMOTEBUILDINGSTATUSREQUEST,
  '__module__' : 'qwak.fitness_service.fitness_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.fitness.service.GetRemoteBuildingStatusRequest)
  })
_sym_db.RegisterMessage(GetRemoteBuildingStatusRequest)

GetRemoteBuildingStatusResponse = _reflection.GeneratedProtocolMessageType('GetRemoteBuildingStatusResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETREMOTEBUILDINGSTATUSRESPONSE,
  '__module__' : 'qwak.fitness_service.fitness_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.fitness.service.GetRemoteBuildingStatusResponse)
  })
_sym_db.RegisterMessage(GetRemoteBuildingStatusResponse)

CleanRemoteBuildResourcesRequest = _reflection.GeneratedProtocolMessageType('CleanRemoteBuildResourcesRequest', (_message.Message,), {
  'DESCRIPTOR' : _CLEANREMOTEBUILDRESOURCESREQUEST,
  '__module__' : 'qwak.fitness_service.fitness_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.fitness.service.CleanRemoteBuildResourcesRequest)
  })
_sym_db.RegisterMessage(CleanRemoteBuildResourcesRequest)

CleanRemoteBuildResourcesResponse = _reflection.GeneratedProtocolMessageType('CleanRemoteBuildResourcesResponse', (_message.Message,), {
  'DESCRIPTOR' : _CLEANREMOTEBUILDRESOURCESRESPONSE,
  '__module__' : 'qwak.fitness_service.fitness_service_pb2'
  # @@protoc_insertion_point(class_scope:qwak.fitness.service.CleanRemoteBuildResourcesResponse)
  })
_sym_db.RegisterMessage(CleanRemoteBuildResourcesResponse)

_FITNESSSERVICE = DESCRIPTOR.services_by_name['FitnessService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\037com.qwak.ai.fitness.service.apiP\001'
  _BUILDMODELREQUEST._serialized_start=139
  _BUILDMODELREQUEST._serialized_end=211
  _BUILDMODELRESPONSE._serialized_start=213
  _BUILDMODELRESPONSE._serialized_end=233
  _CANCELBUILDMODELREQUEST._serialized_start=235
  _CANCELBUILDMODELREQUEST._serialized_end=278
  _CANCELBUILDMODELRESPONSE._serialized_start=280
  _CANCELBUILDMODELRESPONSE._serialized_end=306
  _GETREMOTEBUILDINGSTATUSREQUEST._serialized_start=308
  _GETREMOTEBUILDINGSTATUSREQUEST._serialized_end=358
  _GETREMOTEBUILDINGSTATUSRESPONSE._serialized_start=360
  _GETREMOTEBUILDINGSTATUSRESPONSE._serialized_end=453
  _CLEANREMOTEBUILDRESOURCESREQUEST._serialized_start=455
  _CLEANREMOTEBUILDRESOURCESREQUEST._serialized_end=507
  _CLEANREMOTEBUILDRESOURCESRESPONSE._serialized_start=509
  _CLEANREMOTEBUILDRESOURCESRESPONSE._serialized_end=544
  _FITNESSSERVICE._serialized_start=547
  _FITNESSSERVICE._serialized_end=1055
# @@protoc_insertion_point(module_scope)



from datetime import timedelta
from time import sleep, time

from _qwak_proto.qwak.analytics.analytics_pb2 import (
    QueryResultDownloadURLParams,
    QueryStatus,
)
from _qwak_proto.qwak.analytics.analytics_service_pb2 import (
    GetQueryResultDownloadURLRequest,
    GetQueryResultsPageRequest,
    GetQueryResultsPageResponse,
    QueryRequest,
)
from _qwak_proto.qwak.analytics.analytics_service_pb2_grpc import (
    AnalyticsQueryServiceStub,
)
from dependency_injector.wiring import Provide
from qwak.inner.di_configuration import QwakContainer


class AnalyticsEngineError(RuntimeError):
    def __init__(self, failure_reason: str):
        super().__init__(
            f"Cannot retrieve results from the Qwak Analytics Engine\n{failure_reason}"
        )


class AnalyticsEngineClient:
    def __init__(self, grpc_channel=Provide[QwakContainer.core_grpc_channel]):
        self.grpc_client = AnalyticsQueryServiceStub(grpc_channel)

    def get_analytics_data(self, query: str, timeout: timedelta = None) -> str:
        """
        Sends a given query to the Qwak Analytics Engine and returns the URL of the results file.

        Args:
            query (str): SQL query to be sent to the Qwak Analytics Engine
            timeout (timedelta): maximum time to wait for the query to complete. If None, the function will wait indefinitely.

        Returns:
            str: a path from which the results can be downloaded

        Raises:
            AnalyticsEngineError: if the query failed
            TimeoutError: if the query timed out
        """
        max_time = time() + timeout.total_seconds() if timeout else None
        query_id = self._send_query(query)

        while True:
            status_response = self._check_status(query_id)

            if status_response.status == QueryStatus.SUCCESS:
                download_url = self._get_download_url(query_id)
                return download_url
            elif status_response.status == QueryStatus.FAILED:
                raise AnalyticsEngineError(status_response.failure_reason)
            elif status_response.status == QueryStatus.CANCELED:
                raise AnalyticsEngineError("Query was cancelled")
            elif max_time and time() > max_time:
                raise TimeoutError()
            else:
                sleep(1)

    def _send_query(self, query: str) -> str:
        request = QueryRequest(query=query)
        response = self.grpc_client.Query(request)
        return response.query_id

    def _check_status(self, query_id: str) -> GetQueryResultsPageResponse:
        request = GetQueryResultsPageRequest(
            query_id=query_id, page_id=None, max_results=1
        )
        response = self.grpc_client.GetQueryResultsPage(request)
        return response

    def _get_download_url(self, query_id: str) -> str:
        request_params = QueryResultDownloadURLParams(query_id=query_id)
        request = GetQueryResultDownloadURLRequest(params=request_params)
        response = self.grpc_client.GetQueryResultDownloadURL(request)
        return response.download_url



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import qwak.admiral.user_application_instance.v0.user_application_instance_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class CreateUserApplicationInstanceRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CREATE_OPTIONS_FIELD_NUMBER: builtins.int
    @property
    def create_options(self) -> qwak.admiral.user_application_instance.v0.user_application_instance_pb2.CreateOptions:
        """Options for creation a user application instance"""
    def __init__(
        self,
        *,
        create_options: qwak.admiral.user_application_instance.v0.user_application_instance_pb2.CreateOptions | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["create_options", b"create_options"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["create_options", b"create_options"]) -> None: ...

global___CreateUserApplicationInstanceRequest = CreateUserApplicationInstanceRequest

class CreateUserApplicationInstanceResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___CreateUserApplicationInstanceResponse = CreateUserApplicationInstanceResponse

class UpdateUserApplicationInstanceSpecRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    UPDATE_OPTIONS_FIELD_NUMBER: builtins.int
    @property
    def update_options(self) -> qwak.admiral.user_application_instance.v0.user_application_instance_pb2.UpdateSpecOptions:
        """Options for updating a user application instance"""
    def __init__(
        self,
        *,
        update_options: qwak.admiral.user_application_instance.v0.user_application_instance_pb2.UpdateSpecOptions | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["update_options", b"update_options"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["update_options", b"update_options"]) -> None: ...

global___UpdateUserApplicationInstanceSpecRequest = UpdateUserApplicationInstanceSpecRequest

class UpdateUserApplicationInstanceSpecResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___UpdateUserApplicationInstanceSpecResponse = UpdateUserApplicationInstanceSpecResponse

class GetUserApplicationInstanceRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IDENTIFIER_FIELD_NUMBER: builtins.int
    @property
    def identifier(self) -> qwak.admiral.user_application_instance.v0.user_application_instance_pb2.Identifier:
        """The identifier of the user application instance to get"""
    def __init__(
        self,
        *,
        identifier: qwak.admiral.user_application_instance.v0.user_application_instance_pb2.Identifier | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["identifier", b"identifier"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["identifier", b"identifier"]) -> None: ...

global___GetUserApplicationInstanceRequest = GetUserApplicationInstanceRequest

class GetUserApplicationInstanceResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DESCRIPTION_FIELD_NUMBER: builtins.int
    @property
    def description(self) -> qwak.admiral.user_application_instance.v0.user_application_instance_pb2.Description:
        """Full details of a user application instance"""
    def __init__(
        self,
        *,
        description: qwak.admiral.user_application_instance.v0.user_application_instance_pb2.Description | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["description", b"description"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["description", b"description"]) -> None: ...

global___GetUserApplicationInstanceResponse = GetUserApplicationInstanceResponse

class DeleteUserApplicationInstanceRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IDENTIFIER_FIELD_NUMBER: builtins.int
    @property
    def identifier(self) -> qwak.admiral.user_application_instance.v0.user_application_instance_pb2.Identifier:
        """The identifier of the user application instance to delete"""
    def __init__(
        self,
        *,
        identifier: qwak.admiral.user_application_instance.v0.user_application_instance_pb2.Identifier | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["identifier", b"identifier"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["identifier", b"identifier"]) -> None: ...

global___DeleteUserApplicationInstanceRequest = DeleteUserApplicationInstanceRequest

class DeleteUserApplicationInstanceResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___DeleteUserApplicationInstanceResponse = DeleteUserApplicationInstanceResponse



import itertools
import uuid
from collections import defaultdict
from datetime import datetime

from _qwak_proto.qwak.automation.v1.automation_management_service_pb2 import (
    CreateAutomationRequest,
    CreateAutomationResponse,
    DeleteAutomationRequest,
    DeleteAutomationResponse,
    GetAutomationByNameRequest,
    GetAutomationByNameResponse,
    GetAutomationRequest,
    GetAutomationResponse,
    ListAutomationExecutionsRequest,
    ListAutomationExecutionsResponse,
    ListAutomationsRequest,
    ListAutomationsResponse,
    RegisterAutomationExecutionRequest,
    RegisterAutomationExecutionResponse,
    RunAutomationRequest,
    RunAutomationResponse,
    ToggleAutomationActivityRequest,
    ToggleAutomationActivityResponse,
    UpdateAutomationExecutionRequest,
    UpdateAutomationExecutionResponse,
    UpdateAutomationRequest,
    UpdateAutomationResponse,
)
from _qwak_proto.qwak.automation.v1.automation_management_service_pb2_grpc import (
    AutomationManagementServiceServicer,
)
from _qwak_proto.qwak.automation.v1.automation_pb2 import Automation
from qwak.automations.automation_executions import (
    AutomationExecution,
    ExecutionRunDetails,
    ExecutionStatus,
)
from qwak_services_mock.mocks.utils.exception_handlers import raise_internal_grpc_error


class AutomationManagementServiceMock(AutomationManagementServiceServicer):
    QWAK_ENVIRONMENT = "test_environment"

    def __init__(self):
        self.automations = dict()
        self.automation_executions = defaultdict(list)

    def CreateAutomation(
        self, request: CreateAutomationRequest, context
    ) -> CreateAutomationResponse:
        """Create an automation"""
        try:
            automation_id = str(uuid.uuid4())
            automation = Automation(
                automation_id=automation_id,
                automation_spec=request.automation_spec,
                qwak_environment_id=self.QWAK_ENVIRONMENT,
            )
            self.automations[automation_id] = automation
            return CreateAutomationResponse(automation_id=automation_id)
        except Exception as e:
            raise_internal_grpc_error(context, e)

    def UpdateAutomation(
        self, request: UpdateAutomationRequest, context
    ) -> UpdateAutomationResponse:
        """Update an automation"""
        try:
            automation = Automation(
                automation_id=request.automation_id,
                automation_spec=request.automation_spec,
                qwak_environment_id=self.QWAK_ENVIRONMENT,
            )
            self.automations[request.automation_id] = automation
            return UpdateAutomationResponse()
        except Exception as e:
            raise_internal_grpc_error(context, e)

    def DeleteAutomation(
        self, request: DeleteAutomationRequest, context
    ) -> DeleteAutomationResponse:
        """Delete an automation"""
        try:
            del self.automations[request.automation_id]
            return DeleteAutomationResponse()
        except Exception as e:
            raise_internal_grpc_error(context, e)

    def GetAutomation(
        self, request: GetAutomationRequest, context
    ) -> GetAutomationResponse:
        """Get an automation By ID"""
        try:
            return GetAutomationResponse(
                automation=self.automations[request.automation_id]
            )
        except Exception as e:
            raise_internal_grpc_error(context, e)

    def GetAutomationByName(
        self, request: GetAutomationByNameRequest, context
    ) -> GetAutomationByNameResponse:
        """Get an automation by name"""
        try:
            return GetAutomationByNameResponse(
                automation=list(
                    filter(
                        lambda a: a.automation_spec.automation_name
                        == request.automation_name,
                        self.automations.values(),
                    )
                )[0]
            )
        except Exception as e:
            raise_internal_grpc_error(context, e)

    def ListAutomations(
        self, request: ListAutomationsRequest, context
    ) -> ListAutomationsResponse:
        """List all automation"""
        try:
            return ListAutomationsResponse(automations=list(self.automations.values()))
        except Exception as e:
            raise_internal_grpc_error(context, e)

    def ListAutomationExecutions(
        self, request: ListAutomationExecutionsRequest, context
    ) -> ListAutomationExecutionsResponse:
        """List automation's executions"""
        try:
            return ListAutomationExecutionsResponse(
                automation_executions=self.automation_executions[request.automation_id]
            )
        except Exception as e:
            raise_internal_grpc_error(context, e)

    def RegisterAutomationExecution(
        self, request: RegisterAutomationExecutionRequest, context
    ) -> RegisterAutomationExecutionResponse:
        """Register an execution"""
        try:
            run_details = ExecutionRunDetails(
                start_time=datetime.now(), status=ExecutionStatus.RUNNING
            )
            execution = AutomationExecution(
                execution_id=request.execution_id, run_details=run_details
            )
            self.automation_executions[request.automation_id].append(execution)
            return RegisterAutomationExecutionResponse()
        except Exception as e:
            raise_internal_grpc_error(context, e)

    def UpdateAutomationExecution(
        self, request: UpdateAutomationExecutionRequest, context
    ) -> UpdateAutomationExecutionResponse:
        """Update an execution"""
        try:
            existing_executions = list(
                itertools.chain(*self.automation_executions.values())
            )
            relevant_execution = list(
                filter(
                    lambda ex: ex.execution_id == request.execution_id,
                    existing_executions,
                )
            )[0]
            self.__merge_run_details(
                relevant_execution.run_details, request.run_details
            )
            return UpdateAutomationExecutionResponse()
        except Exception as e:
            raise_internal_grpc_error(context, e)

    def ToggleAutomationActivity(
        self, request: ToggleAutomationActivityRequest, context
    ) -> ToggleAutomationActivityResponse:
        """Toggle whether an automation is enabled or disabled"""
        try:
            self.automations[request.automation_id].automation_spec = request.is_enabled
            return ToggleAutomationActivityResponse()
        except Exception as e:
            raise_internal_grpc_error(context, e)

    def RunAutomation(
        self, request: RunAutomationRequest, context
    ) -> RunAutomationResponse:
        """Run an automation"""
        if self.automations[request.automation_id]:
            return RunAutomationResponse()
        else:
            raise_internal_grpc_error(
                context,
                ValueError(f"Automation with ID {request.automation_id} not found"),
            )

    @staticmethod
    def __merge_run_details(
        run_details_1: ExecutionRunDetails, run_details_2: ExecutionRunDetails
    ):
        run_details_1.start_time = (
            run_details_2.start_time
            if run_details_2.start_time
            else run_details_1.start_time
        )
        run_details_1.end_time = (
            run_details_2.end_time if run_details_2.end_time else run_details_1.end_time
        )
        run_details_1.error_details = (
            run_details_2.error_details
            if run_details_2.error_details
            else run_details_1.error_details
        )
        run_details_1.status = (
            run_details_2.status if run_details_2.status else run_details_1.status
        )
        run_details_1.task = (
            run_details_2.task if run_details_2.task else run_details_1.task
        )
        run_details_1.finish_cause = (
            run_details_2.finish_cause
            if run_details_2.finish_cause
            else run_details_1.finish_cause
        )

        return run_details_1



import logging
from typing import TYPE_CHECKING, List, Optional, Sequence, Tuple

from _qwak_proto.qwak.ecosystem.v0.ecosystem_pb2 import AuthenticatedUnifiedUserContext
from _qwak_proto.qwak.feature_store.serving.serving_pb2 import (
    BatchFeature as ServingProtoBatchFeature,
    BatchV1Feature as ServingProtoBatchV1Feature,
    EntitiesHeader,
    EntityToFeatures,
    EntityValueRow,
    Feature as ServingProtoFeature,
    RequestedEntitiesMatrix,
    RequestedEntitiesMatrixRequest,
    RequestMetaData,
    StreamingAggregationFeature as ServingProtoStreamingAggregationFeature,
    StreamingFeature as ServingProtoStreamingFeature,
    StreamingV1Feature as ServingProtoStreamingV1Feature,
)
from _qwak_proto.qwak.feature_store.serving.serving_pb2_grpc import ServingServiceStub
from qwak.clients.administration.eco_system.client import EcosystemClient
from qwak.exceptions import QwakException
from qwak.feature_store._common.feature_set_utils import (
    BatchFeature,
    BatchFeatureV1,
    StreamingAggregationFeature,
    StreamingFeature,
    StreamingFeatureV1,
)
from qwak.feature_store.online import endpoint_utils
from qwak.feature_store.online.endpoint_utils import EndpointConfig
from qwak.inner.tool.grpc.grpc_tools import create_grpc_channel
from qwak.model.schema import ModelSchema
from qwak.model.schema_entities import BaseFeature, RequestInput

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    try:
        import pandas as pd
    except ImportError:
        pass


class OnlineClient:
    """
    Online Feature Serving client
    """

    def __init__(
        self,
        enable_ssl=True,
        endpoint_url: Optional[str] = None,
        metadata: Optional[Sequence[Tuple[str]]] = None,
    ):
        self._metadata = metadata

        options = (
            ("grpc.keepalive_timeout_ms", 1500),
            ("grpc.client_idle_timeout_ms", 60 * 1000),
        )

        if endpoint_url is None:
            user_context: AuthenticatedUnifiedUserContext = (
                EcosystemClient().get_authenticated_user_context().user
            )
            conf: EndpointConfig = endpoint_utils.get_config_by_account_type(
                user_context
            )
            endpoint_url = conf.endpoint_url
            if conf.enable_ssl is not None:
                enable_ssl = conf.enable_ssl

            if conf.metadata is not None:
                self._metadata = conf.metadata

        channel = create_grpc_channel(
            url=endpoint_url, enable_ssl=enable_ssl, options=options
        )

        self._serving_client: ServingServiceStub = ServingServiceStub(channel)

    @staticmethod
    def to_string_entities_values(values) -> List[str]:
        return [str(value) for value in values]

    def get_feature_values(
        self, schema: ModelSchema, df: "pd.DataFrame", model_name: str = "no-model"
    ) -> "pd.DataFrame":
        """
        :param schema: a ModelSchema object - defines the entities, features and prediction (irelevant in this case).
        :param df: a pandas data-frame with a column for each explicit feature needed
                         and a column for each entity key defined in the schema.
        :return: a pandas data-frame - the feature values defined in the schema
                                       of the requested entities in the df.

        each row in the returned data-frame is constructed by retrieving the most recent requested feature values
        of the entity key(s) for the specific entity value(s) defined in the df.

        TODO: fix imports and align example
        Examples:
        >>> import pandas as pd
        >>> from qwak.feature_store import OnlineClient
        >>> from qwak.model.schema import (
        >>>     ModelSchema, FeatureStoreInput
        >>> )
        >>>
        >>> user_id = Entity(name='uuid', type=str)
        >>>
        >>> model_schema = ModelSchema(
        >>>     entities=[
        >>>         user_id
        >>>     ],
        >>>     inputs=[
        >>>         FeatureStoreInput(entity=user_id, name='user_purchases.number_of_purchases'),
        >>>         FeatureStoreInput(entity=user_id, name='user_purchases.avg_purchase_amount'),
        >>>     ],
        >>>     outputs=[
        >>>         InferenceOutput(name="score", type=float)
        >>>     ])
        >>>
        >>> online_client = OnlineClient()
        >>>
        >>> df = pd.DataFrame(columns=  ['uuid', 'explicit_feature_purchase_price'],
        >>>                   data   =  [ '1'  ,                22                ])
        >>>
        >>> user_features = online_client.get_feature_values(
        >>>                     model_schema,
        >>>                     df)
        >>>
        >>> print(user_features.head())
        >>>	#       user_purchases.number_of_purchases	user_purchases.avg_purchase_amount    otf_quad_price
        >>> # 0	                    76	                              4.796842                     484
        """
        try:
            import pandas as pd
        except ImportError:
            raise QwakException(
                "Missing Pandas dependency required for querying the online feature store"
            )

        (
            entity_features_compounds,
            entities_with_index,
            feature_set_names,
        ) = self._create_entity_and_features_sets(
            schema, pd.DataFrame(df.iloc[0]).transpose()
        )

        if not entity_features_compounds:
            df_result = pd.DataFrame(
                columns=[
                    input_field.name
                    for input_field in schema._inputs
                    if isinstance(input_field, BaseFeature)
                ]
            )

            return pd.concat([df, df_result], axis=1, join="inner")

        entities_to_features = []
        number_of_keys_to_extract = 0
        for entity_features_compound in entity_features_compounds.values():
            number_of_keys_to_extract += len(entity_features_compound.features)
            entities_to_features.append(
                EntityToFeatures(
                    features=entity_features_compound.features,
                    entity_name=entity_features_compound.entity_name,
                )
            )

        ordered_entities_tuple = sorted(
            entities_with_index, key=lambda entity: entity[1]
        )
        ordered_entities = [entity[0] for entity in ordered_entities_tuple]
        entities_values = df[ordered_entities].values.tolist()

        entities_values_req = []
        for index, values in enumerate(entities_values):
            list_of_string_values = self.to_string_entities_values(values)
            entities_values_req.append(
                EntityValueRow(index=index, entity_values=list_of_string_values)
            )

        requested_entities_matrix = RequestedEntitiesMatrix(
            header=EntitiesHeader(entity_names=ordered_entities),
            rows=entities_values_req,
        )

        feature_value_requests = RequestedEntitiesMatrixRequest(
            entity_values_matrix=requested_entities_matrix,
            entities_to_features=entities_to_features,
            request_meta_data=RequestMetaData(
                model_name=model_name,
                feature_set_names=list(feature_set_names),
                num_keys=number_of_keys_to_extract,
            ),
        )

        try:
            response_df_json, _ = self._serving_client.GetMultiFeatures.with_call(
                feature_value_requests, metadata=self._metadata
            )
            result_df = pd.read_json(response_df_json.pandas_df_as_json, orient="split")
            return pd.concat(
                [df.reset_index(drop=True), result_df], axis=1, join="inner"
            )
        except Exception as e:
            raise QwakException(f"Failed to get online features results. Error is: {e}")

    @staticmethod
    def _create_entity_and_features_sets(schema: ModelSchema, df: "pd.DataFrame"):
        entity_features_compounds = {}
        list_of_df_columns = df.columns.to_list()
        entities_and_indexes = []
        feature_set_names = set()
        for entity in schema._entities:
            if entity.name not in df:
                logger.error(
                    f"Schema entity key '{entity.name}' does not exist in the request DataFrame"
                )
            else:
                entity_features_compounds[entity.name] = EntityFeaturesCompound(
                    entity.name, df[entity.name]
                )
                entities_and_indexes.append(
                    (entity.name, list_of_df_columns.index(entity.name))
                )

        for feature in [
            feature
            for feature in schema._inputs
            if not isinstance(feature, RequestInput)
        ]:
            if feature.entity.name not in entity_features_compounds:
                logger.info(
                    f"The entity: {feature.entity.name} of the Feature: {feature} does not exist in the entities list"
                )

            else:
                if isinstance(feature, BatchFeatureV1):
                    feature_proto = ServingProtoFeature(
                        batch_v1_feature=ServingProtoBatchV1Feature(name=feature.name)
                    )
                elif isinstance(feature, BatchFeature):
                    feature_proto = ServingProtoFeature(
                        batch_feature=ServingProtoBatchFeature(name=feature.name)
                    )
                elif isinstance(feature, StreamingFeature):
                    feature_proto = ServingProtoFeature(
                        streaming_feature=ServingProtoStreamingFeature(
                            name=feature.name
                        )
                    )
                elif isinstance(feature, StreamingFeatureV1):
                    feature_proto = ServingProtoFeature(
                        streaming_v1_feature=ServingProtoStreamingV1Feature(
                            name=feature.name
                        )
                    )
                elif isinstance(feature, StreamingAggregationFeature):
                    feature_proto = ServingProtoFeature(
                        streaming_aggregation=ServingProtoStreamingAggregationFeature(
                            name=feature.name
                        )
                    )
                else:
                    raise ValueError(
                        f"Not support type for feature extraction - {type(feature)}"
                    )

                entity_features_compounds[feature.entity.name].add_feature(
                    feature_proto
                )
                feature_set_names.add(feature.name.split(".")[0])

        return entity_features_compounds, entities_and_indexes, feature_set_names


class EntityFeaturesCompound:
    def __init__(self, entity_name, entity_value):
        self.entity_name = entity_name
        self.entity_value = entity_value
        self.features = []

    def add_feature(self, feature):
        self.features.append(feature)



from typing import List, Optional

import grpc
from dependency_injector.wiring import Provide
from _qwak_proto.qwak.builds.build_pb2 import BuildInitiator, RemoteBuildSpec
from _qwak_proto.qwak.builds.internal_builds_orchestrator_service_pb2 import (
    BuildInitDetails,
    InternalBuildModelRequest,
)
from _qwak_proto.qwak.builds.internal_builds_orchestrator_service_pb2_grpc import (
    InternalBuildsOrchestratorServiceStub,
)
from qwak.clients.build_orchestrator.build_model_request_getter import (
    _get_build_model_spec,
)
from qwak.inner.di_configuration import QwakContainer
from qwak.exceptions import QwakException


class InternalBuildOrchestratorClient:
    def __init__(
        self,
        grpc_channel_factory=Provide[
            QwakContainer.internal_grpc_channel_for_builds_orchestrator_factory
        ],
        url_overwrite: Optional[str] = None,
        ssl_overwrite: Optional[bool] = None,
    ):
        self.grpc_channel_factory = grpc_channel_factory
        self._internal_builds_orchestrator_stub: Optional[
            InternalBuildsOrchestratorServiceStub
        ] = None
        self.url_overwrite = url_overwrite
        self.ssl_overwrite = ssl_overwrite

    def build_model(
        self,
        build_conf,
        verbose: int = 3,
        git_commit_id: str = "",
        resolved_model_url: str = "",
        build_code_path: str = "",
        build_v1_flag: bool = False,
        build_config_url: str = "",
        qwak_sdk_wheel_url: str = "",
        qwak_sdk_version_url: str = "",
        build_steps: Optional[List[str]] = None,
        build_initiator: BuildInitiator = None,
        sdk_version: str = "",
    ):
        """Initiate remote build

        Args:
            verbose: log verbosity level
            git_commit_id: commit id
            resolved_model_url: the url of model
            build_conf: the build configuration
            build_code_path: The code  path saved by qwak
            build_v1_flag:
            build_config_url:
            qwak_sdk_wheel_url:
            qwak_sdk_version_url:
            build_steps: List of the steps the build is comprised from
            build_initiator: Override the initiator of build
            sdk_version: The sdk version to build the

        Raises:
            QwakException: In case of failing to connect the service
        """
        if not self._internal_builds_orchestrator_stub:
            grpc_channel = self.grpc_channel_factory(
                self.url_overwrite, self.ssl_overwrite
            )
            if grpc_channel:
                self._internal_builds_orchestrator_stub = (
                    InternalBuildsOrchestratorServiceStub(grpc_channel)
                )

        if not self._internal_builds_orchestrator_stub:
            raise QwakException("Internal GRPC channel is not available.")

        build_steps = build_steps if build_steps else []

        try:
            build_spec: RemoteBuildSpec = _get_build_model_spec(
                build_conf,
                verbose,
                git_commit_id,
                resolved_model_url,
                build_code_path,
                build_v1_flag,
                build_config_url,
                qwak_sdk_wheel_url,
                qwak_sdk_version_url,
                build_steps,
                sdk_version,
            )
            self._internal_builds_orchestrator_stub.BuildModel(
                InternalBuildModelRequest(
                    build_init_details=BuildInitDetails(
                        build_spec=build_spec, initiator=build_initiator
                    )
                )
            )
        except grpc.RpcError as e:
            message = (
                f"Failed to build model, status [{e.code()}] details [{e.details()}]"
            )
            raise QwakException(message)

        except Exception as e:
            message = f"Failed to build model, details [{e}]"
            raise QwakException(message)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qwak/kube_deployment_captain/deployment.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n-qwak/kube_deployment_captain/deployment.proto\x12\x1cqwak.kube.deployment.captain\x1a\x1egoogle/protobuf/wrappers.proto\"\x84\x02\n\x1dKubeAdvancedDeploymentOptions\x12\x19\n\x11number_of_workers\x18\x01 \x01(\x05\x12\x1f\n\x17http_request_timeout_ms\x18\x02 \x01(\x05\x12/\n\x0b\x64\x61\x65mon_mode\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12\x1b\n\x13\x63ustom_iam_role_arn\x18\x04 \x01(\t\x12\x16\n\x0emax_batch_size\x18\x05 \x01(\x05\x12(\n deployment_process_timeout_limit\x18\x06 \x01(\x05\x12\x17\n\x0fpurchase_option\x18\x07 \x01(\t\"\xaf\x01\n\x13KubeServingStrategy\x12G\n\x0frealtime_config\x18\x01 \x01(\x0b\x32,.qwak.kube.deployment.captain.RealTimeConfigH\x00\x12\x43\n\rstream_config\x18\x02 \x01(\x0b\x32*.qwak.kube.deployment.captain.StreamConfigH\x00\x42\n\n\x08Strategy\"\xb5\x01\n\x0f\x41utoScalingSpec\x12\x19\n\x11min_replica_count\x18\x01 \x01(\x05\x12\x19\n\x11max_replica_count\x18\x02 \x01(\x05\x12\x18\n\x10polling_interval\x18\x03 \x01(\x05\x12\x18\n\x10\x63ool_down_period\x18\x04 \x01(\x05\x12\x38\n\x08triggers\x18\x05 \x03(\x0b\x32&.qwak.kube.deployment.captain.Triggers\"i\n\x08Triggers\x12M\n\x12prometheus_trigger\x18\x01 \x01(\x0b\x32/.qwak.kube.deployment.captain.PrometheusTriggerH\x00\x42\x0e\n\x0ctrigger_type\"J\n\x11PrometheusTrigger\x12\x13\n\x0bmetric_name\x18\x01 \x01(\t\x12\x11\n\tthreshold\x18\x02 \x01(\x05\x12\r\n\x05query\x18\x03 \x01(\t\"W\n\x0eRealTimeConfig\x12\'\n\x1fselected_variation_mapping_name\x18\x01 \x01(\t\x12\x1c\n\x14\x61ssumed_iam_role_arn\x18\x02 \x01(\t\"[\n\x0cStreamConfig\x12:\n\x05kafka\x18\x01 \x01(\x0b\x32).qwak.kube.deployment.captain.KafkaConfigH\x00\x42\x0f\n\rconfiguration\"\xb7\x05\n\x0bKafkaConfig\x12\x44\n\x08\x63onsumer\x18\x01 \x01(\x0b\x32\x32.qwak.kube.deployment.captain.KafkaConfig.Consumer\x12\x44\n\x08producer\x18\x02 \x01(\x0b\x32\x32.qwak.kube.deployment.captain.KafkaConfig.Producer\x12\x0f\n\x07workers\x18\x03 \x01(\x05\x1a\x9b\x02\n\x08\x43onsumer\x12\x18\n\x10\x62ootstrap_server\x18\x01 \x03(\t\x12\r\n\x05topic\x18\x02 \x01(\t\x12\r\n\x05group\x18\x03 \x01(\t\x12\x0f\n\x07timeout\x18\x04 \x01(\x05\x12[\n\x10\x61uto_offset_type\x18\x05 \x01(\x0e\x32\x41.qwak.kube.deployment.captain.KafkaConfig.Consumer.AutoOffSetType\x12\x16\n\x0emax_batch_size\x18\x06 \x01(\x05\x12\x18\n\x10max_poll_latency\x18\x07 \x01(\x01\"7\n\x0e\x41utoOffSetType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0c\n\x08\x45\x41RLIEST\x10\x01\x12\n\n\x06LATEST\x10\x02\x1a\xec\x01\n\x08Producer\x12\x18\n\x10\x62ootstrap_server\x18\x01 \x03(\t\x12\r\n\x05topic\x18\x02 \x01(\t\x12\\\n\x10\x63ompression_type\x18\x03 \x01(\x0e\x32\x42.qwak.kube.deployment.captain.KafkaConfig.Producer.CompressionType\"Y\n\x0f\x43ompressionType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x10\n\x0cUNCOMPRESSED\x10\x01\x12\x08\n\x04GZIP\x10\x02\x12\n\n\x06SNAPPY\x10\x03\x12\x07\n\x03LZ4\x10\x04\x12\x08\n\x04ZSTD\x10\x05\"\xa9\x05\n\x12KubeDeploymentSize\x12\x16\n\x0enumber_of_pods\x18\x01 \x01(\x05\x12\x0b\n\x03\x63pu\x18\x02 \x01(\x02\x12\x15\n\rmemory_amount\x18\x03 \x01(\x05\x12Q\n\x0cmemory_units\x18\x04 \x01(\x0e\x32;.qwak.kube.deployment.captain.KubeDeploymentSize.MemoryUnit\x12T\n\rgpu_resources\x18\x05 \x01(\x0b\x32=.qwak.kube.deployment.captain.KubeDeploymentSize.GpuResources\x1an\n\x0cGpuResources\x12J\n\x08gpu_type\x18\x01 \x01(\x0e\x32\x38.qwak.kube.deployment.captain.KubeDeploymentSize.GpuType\x12\x12\n\ngpu_amount\x18\x02 \x01(\x05\"\x90\x02\n\x07GpuType\x12\x0f\n\x0bINVALID_GPU\x10\x00\x12\x0e\n\nNVIDIA_K80\x10\x01\x12\x0f\n\x0bNVIDIA_V100\x10\x02\x12\x0f\n\x0bNVIDIA_A100\x10\x03\x12\r\n\tNVIDIA_T4\x10\x04\x12\x0f\n\x0bNVIDIA_A10G\x10\x05\x12\r\n\tNVIDIA_L4\x10\x06\x12\x14\n\x10NVIDIA_T4_1_4_15\x10\x07\x12\x14\n\x10NVIDIA_T4_1_8_30\x10\x08\x12\x15\n\x11NVIDIA_T4_1_16_60\x10\t\x12\x1e\n\x1aNVIDIA_A100_80GB_8_96_1360\x10\n\x12\x16\n\x12NVIDIA_V100_1_8_52\x10\x0b\x12\x18\n\x14NVIDIA_V100_4_32_208\x10\x0c\"+\n\nMemoryUnit\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x07\n\x03MIB\x10\x01\x12\x07\n\x03GIB\x10\x02\"\xc3\x01\n\x1dKubeRuntimeDeploymentSettings\x12_\n\x11root_logger_level\x18\x01 \x01(\x0e\x32\x44.qwak.kube.deployment.captain.KubeRuntimeDeploymentSettings.LogLevel\"A\n\x08LogLevel\x12\x0b\n\x07NOT_SET\x10\x00\x12\t\n\x05\x45RROR\x10\x01\x12\x08\n\x04WARN\x10\x02\x12\x08\n\x04INFO\x10\x03\x12\t\n\x05\x44\x45\x42UG\x10\x04*;\n\x12KubeDeploymentType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0c\n\x08REALTIME\x10\x01\x12\n\n\x06STREAM\x10\x02\x42+\n\'com.qwak.ai.kube.deployment.captain.apiP\x01\x62\x06proto3')

_KUBEDEPLOYMENTTYPE = DESCRIPTOR.enum_types_by_name['KubeDeploymentType']
KubeDeploymentType = enum_type_wrapper.EnumTypeWrapper(_KUBEDEPLOYMENTTYPE)
UNKNOWN = 0
REALTIME = 1
STREAM = 2


_KUBEADVANCEDDEPLOYMENTOPTIONS = DESCRIPTOR.message_types_by_name['KubeAdvancedDeploymentOptions']
_KUBESERVINGSTRATEGY = DESCRIPTOR.message_types_by_name['KubeServingStrategy']
_AUTOSCALINGSPEC = DESCRIPTOR.message_types_by_name['AutoScalingSpec']
_TRIGGERS = DESCRIPTOR.message_types_by_name['Triggers']
_PROMETHEUSTRIGGER = DESCRIPTOR.message_types_by_name['PrometheusTrigger']
_REALTIMECONFIG = DESCRIPTOR.message_types_by_name['RealTimeConfig']
_STREAMCONFIG = DESCRIPTOR.message_types_by_name['StreamConfig']
_KAFKACONFIG = DESCRIPTOR.message_types_by_name['KafkaConfig']
_KAFKACONFIG_CONSUMER = _KAFKACONFIG.nested_types_by_name['Consumer']
_KAFKACONFIG_PRODUCER = _KAFKACONFIG.nested_types_by_name['Producer']
_KUBEDEPLOYMENTSIZE = DESCRIPTOR.message_types_by_name['KubeDeploymentSize']
_KUBEDEPLOYMENTSIZE_GPURESOURCES = _KUBEDEPLOYMENTSIZE.nested_types_by_name['GpuResources']
_KUBERUNTIMEDEPLOYMENTSETTINGS = DESCRIPTOR.message_types_by_name['KubeRuntimeDeploymentSettings']
_KAFKACONFIG_CONSUMER_AUTOOFFSETTYPE = _KAFKACONFIG_CONSUMER.enum_types_by_name['AutoOffSetType']
_KAFKACONFIG_PRODUCER_COMPRESSIONTYPE = _KAFKACONFIG_PRODUCER.enum_types_by_name['CompressionType']
_KUBEDEPLOYMENTSIZE_GPUTYPE = _KUBEDEPLOYMENTSIZE.enum_types_by_name['GpuType']
_KUBEDEPLOYMENTSIZE_MEMORYUNIT = _KUBEDEPLOYMENTSIZE.enum_types_by_name['MemoryUnit']
_KUBERUNTIMEDEPLOYMENTSETTINGS_LOGLEVEL = _KUBERUNTIMEDEPLOYMENTSETTINGS.enum_types_by_name['LogLevel']
KubeAdvancedDeploymentOptions = _reflection.GeneratedProtocolMessageType('KubeAdvancedDeploymentOptions', (_message.Message,), {
  'DESCRIPTOR' : _KUBEADVANCEDDEPLOYMENTOPTIONS,
  '__module__' : 'qwak.kube_deployment_captain.deployment_pb2'
  # @@protoc_insertion_point(class_scope:qwak.kube.deployment.captain.KubeAdvancedDeploymentOptions)
  })
_sym_db.RegisterMessage(KubeAdvancedDeploymentOptions)

KubeServingStrategy = _reflection.GeneratedProtocolMessageType('KubeServingStrategy', (_message.Message,), {
  'DESCRIPTOR' : _KUBESERVINGSTRATEGY,
  '__module__' : 'qwak.kube_deployment_captain.deployment_pb2'
  # @@protoc_insertion_point(class_scope:qwak.kube.deployment.captain.KubeServingStrategy)
  })
_sym_db.RegisterMessage(KubeServingStrategy)

AutoScalingSpec = _reflection.GeneratedProtocolMessageType('AutoScalingSpec', (_message.Message,), {
  'DESCRIPTOR' : _AUTOSCALINGSPEC,
  '__module__' : 'qwak.kube_deployment_captain.deployment_pb2'
  # @@protoc_insertion_point(class_scope:qwak.kube.deployment.captain.AutoScalingSpec)
  })
_sym_db.RegisterMessage(AutoScalingSpec)

Triggers = _reflection.GeneratedProtocolMessageType('Triggers', (_message.Message,), {
  'DESCRIPTOR' : _TRIGGERS,
  '__module__' : 'qwak.kube_deployment_captain.deployment_pb2'
  # @@protoc_insertion_point(class_scope:qwak.kube.deployment.captain.Triggers)
  })
_sym_db.RegisterMessage(Triggers)

PrometheusTrigger = _reflection.GeneratedProtocolMessageType('PrometheusTrigger', (_message.Message,), {
  'DESCRIPTOR' : _PROMETHEUSTRIGGER,
  '__module__' : 'qwak.kube_deployment_captain.deployment_pb2'
  # @@protoc_insertion_point(class_scope:qwak.kube.deployment.captain.PrometheusTrigger)
  })
_sym_db.RegisterMessage(PrometheusTrigger)

RealTimeConfig = _reflection.GeneratedProtocolMessageType('RealTimeConfig', (_message.Message,), {
  'DESCRIPTOR' : _REALTIMECONFIG,
  '__module__' : 'qwak.kube_deployment_captain.deployment_pb2'
  # @@protoc_insertion_point(class_scope:qwak.kube.deployment.captain.RealTimeConfig)
  })
_sym_db.RegisterMessage(RealTimeConfig)

StreamConfig = _reflection.GeneratedProtocolMessageType('StreamConfig', (_message.Message,), {
  'DESCRIPTOR' : _STREAMCONFIG,
  '__module__' : 'qwak.kube_deployment_captain.deployment_pb2'
  # @@protoc_insertion_point(class_scope:qwak.kube.deployment.captain.StreamConfig)
  })
_sym_db.RegisterMessage(StreamConfig)

KafkaConfig = _reflection.GeneratedProtocolMessageType('KafkaConfig', (_message.Message,), {

  'Consumer' : _reflection.GeneratedProtocolMessageType('Consumer', (_message.Message,), {
    'DESCRIPTOR' : _KAFKACONFIG_CONSUMER,
    '__module__' : 'qwak.kube_deployment_captain.deployment_pb2'
    # @@protoc_insertion_point(class_scope:qwak.kube.deployment.captain.KafkaConfig.Consumer)
    })
  ,

  'Producer' : _reflection.GeneratedProtocolMessageType('Producer', (_message.Message,), {
    'DESCRIPTOR' : _KAFKACONFIG_PRODUCER,
    '__module__' : 'qwak.kube_deployment_captain.deployment_pb2'
    # @@protoc_insertion_point(class_scope:qwak.kube.deployment.captain.KafkaConfig.Producer)
    })
  ,
  'DESCRIPTOR' : _KAFKACONFIG,
  '__module__' : 'qwak.kube_deployment_captain.deployment_pb2'
  # @@protoc_insertion_point(class_scope:qwak.kube.deployment.captain.KafkaConfig)
  })
_sym_db.RegisterMessage(KafkaConfig)
_sym_db.RegisterMessage(KafkaConfig.Consumer)
_sym_db.RegisterMessage(KafkaConfig.Producer)

KubeDeploymentSize = _reflection.GeneratedProtocolMessageType('KubeDeploymentSize', (_message.Message,), {

  'GpuResources' : _reflection.GeneratedProtocolMessageType('GpuResources', (_message.Message,), {
    'DESCRIPTOR' : _KUBEDEPLOYMENTSIZE_GPURESOURCES,
    '__module__' : 'qwak.kube_deployment_captain.deployment_pb2'
    # @@protoc_insertion_point(class_scope:qwak.kube.deployment.captain.KubeDeploymentSize.GpuResources)
    })
  ,
  'DESCRIPTOR' : _KUBEDEPLOYMENTSIZE,
  '__module__' : 'qwak.kube_deployment_captain.deployment_pb2'
  # @@protoc_insertion_point(class_scope:qwak.kube.deployment.captain.KubeDeploymentSize)
  })
_sym_db.RegisterMessage(KubeDeploymentSize)
_sym_db.RegisterMessage(KubeDeploymentSize.GpuResources)

KubeRuntimeDeploymentSettings = _reflection.GeneratedProtocolMessageType('KubeRuntimeDeploymentSettings', (_message.Message,), {
  'DESCRIPTOR' : _KUBERUNTIMEDEPLOYMENTSETTINGS,
  '__module__' : 'qwak.kube_deployment_captain.deployment_pb2'
  # @@protoc_insertion_point(class_scope:qwak.kube.deployment.captain.KubeRuntimeDeploymentSettings)
  })
_sym_db.RegisterMessage(KubeRuntimeDeploymentSettings)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\'com.qwak.ai.kube.deployment.captain.apiP\001'
  _KUBEDEPLOYMENTTYPE._serialized_start=2681
  _KUBEDEPLOYMENTTYPE._serialized_end=2740
  _KUBEADVANCEDDEPLOYMENTOPTIONS._serialized_start=112
  _KUBEADVANCEDDEPLOYMENTOPTIONS._serialized_end=372
  _KUBESERVINGSTRATEGY._serialized_start=375
  _KUBESERVINGSTRATEGY._serialized_end=550
  _AUTOSCALINGSPEC._serialized_start=553
  _AUTOSCALINGSPEC._serialized_end=734
  _TRIGGERS._serialized_start=736
  _TRIGGERS._serialized_end=841
  _PROMETHEUSTRIGGER._serialized_start=843
  _PROMETHEUSTRIGGER._serialized_end=917
  _REALTIMECONFIG._serialized_start=919
  _REALTIMECONFIG._serialized_end=1006
  _STREAMCONFIG._serialized_start=1008
  _STREAMCONFIG._serialized_end=1099
  _KAFKACONFIG._serialized_start=1102
  _KAFKACONFIG._serialized_end=1797
  _KAFKACONFIG_CONSUMER._serialized_start=1275
  _KAFKACONFIG_CONSUMER._serialized_end=1558
  _KAFKACONFIG_CONSUMER_AUTOOFFSETTYPE._serialized_start=1503
  _KAFKACONFIG_CONSUMER_AUTOOFFSETTYPE._serialized_end=1558
  _KAFKACONFIG_PRODUCER._serialized_start=1561
  _KAFKACONFIG_PRODUCER._serialized_end=1797
  _KAFKACONFIG_PRODUCER_COMPRESSIONTYPE._serialized_start=1708
  _KAFKACONFIG_PRODUCER_COMPRESSIONTYPE._serialized_end=1797
  _KUBEDEPLOYMENTSIZE._serialized_start=1800
  _KUBEDEPLOYMENTSIZE._serialized_end=2481
  _KUBEDEPLOYMENTSIZE_GPURESOURCES._serialized_start=2051
  _KUBEDEPLOYMENTSIZE_GPURESOURCES._serialized_end=2161
  _KUBEDEPLOYMENTSIZE_GPUTYPE._serialized_start=2164
  _KUBEDEPLOYMENTSIZE_GPUTYPE._serialized_end=2436
  _KUBEDEPLOYMENTSIZE_MEMORYUNIT._serialized_start=2438
  _KUBEDEPLOYMENTSIZE_MEMORYUNIT._serialized_end=2481
  _KUBERUNTIMEDEPLOYMENTSETTINGS._serialized_start=2484
  _KUBERUNTIMEDEPLOYMENTSETTINGS._serialized_end=2679
  _KUBERUNTIMEDEPLOYMENTSETTINGS_LOGLEVEL._serialized_start=2614
  _KUBERUNTIMEDEPLOYMENTSETTINGS_LOGLEVEL._serialized_end=2679
# @@protoc_insertion_point(module_scope)



import functools
import inspect
from dataclasses import dataclass
from datetime import datetime
from inspect import Signature
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

from qwak.feature_store._common.artifact_utils import ArtifactSpec, ArtifactsUploader
from qwak.feature_store._common.feature_set_utils import get_batch_source_for_featureset
from qwak.feature_store.feature_sets._utils._featureset_utils import FeaturesetUtils
from qwak.feature_store.feature_sets.base_feature_set import BaseFeatureSet
from qwak.feature_store.validations.validation_options import (
    FeatureSetValidationOptions,
)

if TYPE_CHECKING:
    try:
        import pandas as pd
    except ImportError:
        pass

from _qwak_proto.qwak.feature_store.features.execution_pb2 import (
    ExecutionSpec as ProtoExecutionSpec,
)
from _qwak_proto.qwak.feature_store.features.feature_set_pb2 import FeatureSetSpec
from _qwak_proto.qwak.feature_store.features.feature_set_types_pb2 import (
    BatchFeatureSetV1 as ProtoBatchFeatureSetV1,
    FeatureSetBatchSource as ProtoFeatureSetBatchSource,
    FeatureSetType as ProtoFeatureSetType,
)
from qwak.feature_store.entities.entity import Entity
from qwak.feature_store.feature_sets.backfill import Backfill
from qwak.feature_store.feature_sets.context import Context
from qwak.feature_store.feature_sets.execution_spec import ClusterTemplate
from qwak.feature_store.feature_sets.metadata import (
    Metadata,
    get_metadata_from_function,
    set_metadata_on_function,
)
from qwak.feature_store.feature_sets.read_policies import ReadPolicy, ReadPolicyABC
from qwak.feature_store.feature_sets.transformations import BaseTransformation
from typeguard import typechecked

# decorator attributes
_BACKFILL_POLICY_ATTRIBUTE = "_qwak_backfill_policy"
_EXECUTION_SPECIFICATION_ATTRIBUTE = "_qwak_execution_specification"
_SCHEDULING_POLICY_ATTRIBUTE = "_qwak_scheduling_policy"

# Default timestamp column name for TimeFrame
_DEFAULT_TIMEFRAME_TS_COL_NAME = "qwak_window_end_ts"


def feature_set(
    *,
    data_sources: Union[Dict[str, ReadPolicyABC], List[str]],
    timestamp_column_name: str = None,
    name: str = None,
    entity: Optional[str] = None,
    key: Optional[str] = None,
):
    """
    Define a batch scheduled feature set. Default scheduling policy is every 4 hours.

    :param name: The name of the feature set. If not defined, taken from the function name
    :param entity: a string reference to a Qwak entity, or the entity definition itself.
                  Only one of, entity or key, can be provided.
    :param key: a column name in the feature set which is the key. Only one of, entity or key, can be provided.
    :param data_sources: a string reference to a Qwak data source, or the data source definition itself
    :param timestamp_column_name: Timestamp column the feature set should consider when ingesting records.
    If more than one data source is specified - the field is mandatory. If only one data source is specified - the
    field will be taken by from the data source definition.

    Example:

    ... code-block:: python
        @batch.feature_set(entity="users", data_sources=["snowflake_users_table"])
        def user_features():
            return SparkSqlTransformation("SELECT user_id, age FROM data_source")

    Example with key:

    ... code-block:: python
    @batch.feature_set(key="user_id", data_sources=["snowflake_users_table"])
    def user_features():
        return SparkSqlTransformation("SELECT user_id, age FROM data_source")
    """

    def decorator(function):
        sig: Signature = inspect.signature(function)
        if "context" in sig.parameters:
            user_transformation = function(context=Context())
        else:
            user_transformation = function()

        FeaturesetUtils.validate_base_featureset_decorator(
            user_transformation=user_transformation, entity=entity, key=key
        )

        fs_name = name or function.__name__
        batch_feature_set = BatchFeatureSet(
            name=fs_name,
            entity=entity if entity else None,
            key=key if key else None,
            data_sources=data_sources,
            timestamp_column_name=timestamp_column_name,
            transformation=user_transformation,
            scheduling_policy=getattr(
                function, _SCHEDULING_POLICY_ATTRIBUTE, "0 */4 * * *"
            ),
            metadata=get_metadata_from_function(
                function, description=fs_name, display_name=fs_name
            ),
            cluster_template=getattr(
                function,
                _EXECUTION_SPECIFICATION_ATTRIBUTE,
                ClusterTemplate.MEDIUM,
            ),
            backfill=getattr(function, _BACKFILL_POLICY_ATTRIBUTE, None),
            __instance_module_path__=inspect.stack()[1].filename,
        )

        functools.update_wrapper(batch_feature_set, user_transformation)
        return batch_feature_set

    return decorator


@typechecked
def scheduling(*, cron_expression: Optional[str]):
    """
    Sets the scheduling policy of the batch feature set according to the given cron expression
    A cron expression is a string consisting of six or seven subexpressions (fields) that describe individual details
    of the schedule. These fields, separated by white space, can contain any of the allowed values with various
    combinations of the allowed characters for that field.

    If None was passed instead of cron expression, the feature set will not be scheduled,
    and would be triggered only by user request

    Some more examples:

    Expression	Means
    0 0 12 * * ?	Fire at 12:00 PM (noon) every day
    0 15 10 ? * *	Fire at 10:15 AM every day
    0 15 10 * * ?	Fire at 10:15 AM every day
    0 15 10 * * ? *	Fire at 10:15 AM every day
    0 15 10 * * ? 2005	Fire at 10:15 AM every day during the year 2005
    0 * 14 * * ?	Fire every minute starting at 2:00 PM and ending at 2:59 PM, every day

    :param cron_expression: cron expression

    Example:

    ... code-block:: python

        @batch.feature_set(entity="users", data_sources=["snowflake_users_table"])
        @batch.scheduling(cron_expression="0 8 * * *")
        def user_features():
            return SparkSqlTransformation("SELECT user_id, age FROM data_source")

    """

    # TODO: add backend validation on cron expression
    def decorator(user_transformation):
        _validate_decorator_ordering(user_transformation)
        setattr(user_transformation, _SCHEDULING_POLICY_ATTRIBUTE, cron_expression)

        return user_transformation

    return decorator


@typechecked
def metadata(
    *,
    owner: Optional[str] = None,
    description: Optional[str] = None,
    display_name: Optional[str] = None,
):
    """
    Sets additional user provided metadata

    :param owner: feature set owner
    :param description: General description of the feature set
    :param display_name: Human readable name of the feature set

    Example:

    ... code-block:: python

        @batch.feature_set(entity="users", data_sources=["snowflake_users_table"])
        @batch.metadata(
            owner="datainfra@qwak.com",
            display_name="User Batch Features",
            description="Users feature from the Snowflake replica of the production users table",
        )
        def user_features():
            return SparkSqlTransformation("SELECT user_id, age FROM data_source")

    """

    def decorator(user_transformation):
        _validate_decorator_ordering(user_transformation)
        set_metadata_on_function(
            function=user_transformation,
            owner=owner,
            description=description,
            display_name=display_name,
        )

        return user_transformation

    return decorator


@typechecked
def backfill(
    *,
    start_date: datetime,
):
    """
    Set the backfill policy of the feature set.

    :param start_date: Start date of the backfill process. Data is extracted starting from this date
    :param end_date: Optionally the end date of the backfill process. Cuts off the backfill according to this date

    Example:

    ... code-block:: python
        @batch.feature_set(entity="users", data_sources=["snowflake_users_table"])
        @batch.backfill(start_date=datetime(2022, 1, 1))
        def user_features():
            return SparkSqlTransformation("SELECT user_id, age FROM data_source")

    """

    def decorator(user_transformation):
        _validate_decorator_ordering(user_transformation)
        setattr(
            user_transformation,
            _BACKFILL_POLICY_ATTRIBUTE,
            Backfill(start_date=start_date),
        )

        return user_transformation

    return decorator


@typechecked
def execution_specification(
    *,
    cluster_template: ClusterTemplate,
):
    """
    Set the execution specification of the cluster running the feature set

    :param cluster_template: Predefined template sizes

    Cluster template example:

    ... code-block:: python
        @batch.feature_set(entity="users", data_sources=["snowflake_users_table"])
        @batch.execution_specification(cluster_template=ClusterTemplate.MEDIUM)
        def user_features():
            return SparkSqlTransformation("SELECT user_id, age FROM data_source"

    """

    def decorator(user_transformation):
        _validate_decorator_ordering(user_transformation)
        setattr(
            user_transformation, _EXECUTION_SPECIFICATION_ATTRIBUTE, cluster_template
        )

        return user_transformation

    return decorator


def _validate_decorator_ordering(user_transformation):
    if isinstance(user_transformation, BatchFeatureSet):
        raise ValueError(
            "Wrong decorator ordering - @batch.feature_set should be the top most decorator"
        )


@dataclass
class BatchFeatureSet(BaseFeatureSet):
    timestamp_column_name: str = str()
    scheduling_policy: str = str()
    transformation: Optional[BaseTransformation] = None
    metadata: Optional[Metadata] = None
    cluster_template: Optional[ClusterTemplate] = None
    backfill: Optional[Backfill] = None

    def __post_init__(self):
        self._validate()

    @classmethod
    def _from_proto(cls, proto: FeatureSetSpec):
        batch_v1_def = proto.feature_set_type.batch_feature_set_v1

        return cls(
            name=proto.name,
            entity=Entity._from_proto(proto.entity),
            data_sources=[
                ds.data_source.name for ds in batch_v1_def.feature_set_batch_sources
            ],
            timestamp_column_name=batch_v1_def.timestamp_column_name,
            scheduling_policy=batch_v1_def.scheduling_policy,
            transformation=BaseTransformation._from_proto(batch_v1_def.transformation),
            metadata=Metadata.from_proto(proto.metadata),
            cluster_template=ClusterTemplate.from_proto(batch_v1_def.execution_spec),
            backfill=Backfill.from_proto(batch_v1_def.backfill),
        )

    def _get_data_sources(self, feature_registry) -> List[ProtoFeatureSetBatchSource]:
        if isinstance(self.data_sources, list):
            actual_data_sources = {
                ds_name: ReadPolicy.NewOnly for ds_name in self.data_sources
            }
        else:
            actual_data_sources = self.data_sources

        return [
            ProtoFeatureSetBatchSource(
                data_source=get_batch_source_for_featureset(ds, feature_registry),
                read_policy=(
                    read_policy().to_proto()
                    if inspect.isclass(read_policy)
                    else read_policy.to_proto()
                ),
            )
            for ds, read_policy in actual_data_sources.items()
        ]

    def _get_timestamp_column(
        self, data_sources: List[ProtoFeatureSetBatchSource]
    ) -> str:
        if self.timestamp_column_name:
            return self.timestamp_column_name

        # if no explicit timestamp column was set AND there's a TimeFrame
        # datasource, set the default column name.
        has_timeframe: bool = any(
            [ds.read_policy.WhichOneof("type") == "time_frame" for ds in data_sources]
        )
        if has_timeframe:
            return _DEFAULT_TIMEFRAME_TS_COL_NAME

        if len(data_sources) >= 2:
            raise ValueError(
                "If more than one data source is defined - `timestamp_column_name` on the feature set decorator must be defined"
            )

        # if we got here - only one data source is defined
        return data_sources[0].data_source.date_created_column

    def _validate(self):
        super()._validate()

    def _to_proto(
        self,
        git_commit,
        features,
        feature_registry,
        artifact_url: Optional[str] = str(),
        **kwargs,
    ) -> Tuple[FeatureSetSpec, Optional[str]]:
        self._validate()

        data_sources = self._get_data_sources(feature_registry)

        if not artifact_url:
            artifact: Optional[ArtifactSpec] = ArtifactsUploader.get_artifact_spec(
                self.transformation, self.name, self.__instance_module_path__
            )
            if artifact:
                artifact_url = ArtifactsUploader.upload(artifact)

        return (
            FeatureSetSpec(
                name=self.name,
                metadata=self.metadata.to_proto(),
                git_commit=git_commit,
                features=features,
                entity=self._get_entity_definition(feature_registry),
                feature_set_type=ProtoFeatureSetType(
                    batch_feature_set_v1=ProtoBatchFeatureSetV1(
                        online_sink=True,
                        offline_sink=True,
                        timestamp_column_name=self._get_timestamp_column(data_sources),
                        scheduling_policy=self.scheduling_policy,
                        feature_set_batch_sources=data_sources,
                        execution_spec=ProtoExecutionSpec(
                            cluster_template=(
                                ClusterTemplate.to_proto(self.cluster_template)
                                if self.cluster_template
                                else None
                            )
                        ),
                        transformation=(
                            self.transformation._to_proto(artifact_path=artifact_url)
                            if self.transformation
                            else None
                        ),
                        backfill=self.backfill.to_proto() if self.backfill else None,
                    )
                ),
            ),
            artifact_url,
        )

    def get_sample(
        self,
        number_of_rows: int = 10,
        validation_options: Optional[FeatureSetValidationOptions] = None,
    ) -> "pd.DataFrame":
        """
        Fetches a sample of the Feature set transformation by loading requested sample of data from the data source
        and executing the transformation on that data.

        :param number_of_rows: number of rows requests
        :param validation_options: validation options
        :returns Sample Pandas Dataframe

        Example with entity:

        ... code-block:: python
            @batch.feature_set(entity="users", data_sources=["snowflake_users_table"])
            @batch.backfill(start_date=datetime(2022, 1, 1))
            def user_features():
                return SparkSqlTransformation("SELECT user_id, age FROM data_source")

            sample_features = user_features.get_sample()
            print(sample_feature)
            #	    user_id	         timestamp	        user_features.age
            # 0	      1	        2021-01-02 17:00:00	              23
            # 1	      1	        2021-01-01 12:00:00	              51
            # 2	      2	        2021-01-02 12:00:00	              66
            # 3	      2	        2021-01-01 18:00:00	              34

        Example with key:

        ... code-block:: python
            @batch.feature_set(key="user_id", data_sources=["snowflake_users_table"])
            @batch.backfill(start_date=datetime(2022, 1, 1))
            def user_features():
                return SparkSqlTransformation("SELECT user_id, age FROM data_source")

            sample_features = user_features.get_sample()
            print(sample_feature)
            #	    user_id	         timestamp	        user_features.age
            # 0	      1	        2021-01-02 17:00:00	              23
            # 1	      1	        2021-01-01 12:00:00	              51
            # 2	      2	        2021-01-02 12:00:00	              66
            # 3	      2	        2021-01-01 18:00:00	              34
        """
        return super().get_sample(
            number_of_rows=number_of_rows, validation_options=validation_options
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qwak/administration/account/v1/personalization.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n4qwak/administration/account/v1/personalization.proto\x12\x1bqwak.administration.account\"K\n\x16\x41\x63\x63ountPersonalization\x12\x31\n\x05\x62rand\x18\x01 \x01(\x0b\x32\".qwak.administration.account.Brand\" \n\x05\x42rand\x12\x17\n\x0flogo_image_path\x18\x01 \x01(\tB\xb4\x01\n)com.qwak.ai.administration.api.account.v1P\x01Z\x84\x01github.com/qwak-ai/qwak-platform/services/core/java/user-management/user-management-api/pb/qwak/administration/account/v1;account_v0b\x06proto3')



_ACCOUNTPERSONALIZATION = DESCRIPTOR.message_types_by_name['AccountPersonalization']
_BRAND = DESCRIPTOR.message_types_by_name['Brand']
AccountPersonalization = _reflection.GeneratedProtocolMessageType('AccountPersonalization', (_message.Message,), {
  'DESCRIPTOR' : _ACCOUNTPERSONALIZATION,
  '__module__' : 'qwak.administration.account.v1.personalization_pb2'
  # @@protoc_insertion_point(class_scope:qwak.administration.account.AccountPersonalization)
  })
_sym_db.RegisterMessage(AccountPersonalization)

Brand = _reflection.GeneratedProtocolMessageType('Brand', (_message.Message,), {
  'DESCRIPTOR' : _BRAND,
  '__module__' : 'qwak.administration.account.v1.personalization_pb2'
  # @@protoc_insertion_point(class_scope:qwak.administration.account.Brand)
  })
_sym_db.RegisterMessage(Brand)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n)com.qwak.ai.administration.api.account.v1P\001Z\204\001github.com/qwak-ai/qwak-platform/services/core/java/user-management/user-management-api/pb/qwak/administration/account/v1;account_v0'
  _ACCOUNTPERSONALIZATION._serialized_start=85
  _ACCOUNTPERSONALIZATION._serialized_end=160
  _BRAND._serialized_start=162
  _BRAND._serialized_end=194
# @@protoc_insertion_point(module_scope)
